# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class HeapElement(object):
    """ An element of one of the heaps on SpiNNaker.
    """

    __slots__ = [
        # A pointer to the block
        "_block_address",
        # A pointer to the next block
        "_next_address",
        # True if the block is free
        "_is_free",
        # The tag of the block
        "_tag",
        # The app ID of the block
        "_app_id"
    ]

    def __init__(self, block_address, next_address, free):
        """
        :param int block_address: The address of this element on the heap
        :param int next_address: The address of the next element on the heap
        :param int free: The "free" element of the block as read from the heap
        """
        self._block_address = block_address
        self._next_address = next_address
        self._is_free = (free & 0xFFFF0000) != 0xFFFF0000
        self._tag = None
        self._app_id = None
        if not self._is_free:
            self._tag = free & 0xFF
            self._app_id = (free >> 8) & 0xFF

    @property
    def block_address(self):
        """ The address of the block

        :rtype: int
        """
        return self._block_address

    @property
    def next_address(self):
        """ The address of the next block, or 0 if none

        :rtype: int
        """
        return self._next_address

    @property
    def size(self):
        """ The usable size of this block (not including the header)

        :rtype: int
        """
        return self._next_address - self._block_address - 8

    @property
    def is_free(self):
        """ True if this block is a free block, False otherwise

        :rtype: bool
        """
        return self._is_free

    @property
    def tag(self):
        """ The tag of the block if allocated, or None if not

        :rtype: int or None
        """
        return self._tag

    @property
    def app_id(self):
        """ The application ID of the block if allocated, or None if not

        :rtype: int or None
        """
        return self._app_id

    def __str__(self):
        if self._is_free:
            return "FREE  0x{:8X} SIZE: {:9d}".format(
                self._block_address, self.size)
        return "BLOCK 0x{:8X} SIZE: {:9d} TAG: {:3d} APP_ID: {:3d}".format(
            self._block_address, self.size, self._tag, self._app_id)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" This is a script used to check the state of a SpiNNaker machine.
"""

import sys
import argparse
from spinnman.transceiver import create_transceiver_from_hostname
from spinn_machine import CoreSubsets, CoreSubset
from spinnman.model.enums import CPUState

SCAMP_ID = 0
IGNORED_IDS = {SCAMP_ID, 16}  # WHY 16?


def get_cores_in_run_state(txrx, app_id, print_all_chips):
    """
    :param Transceiver txrx:
    :param int app_id:
    :param bool print_all_chips:
    """
    count_finished = txrx.get_core_state_count(app_id, CPUState.FINISHED)
    count_run = txrx.get_core_state_count(app_id, CPUState.RUNNING)
    print('running: {} finished: {}'.format(count_run, count_finished))

    machine = txrx.get_machine_details()
    print('machine max x: {} max y: {}'.format(
        machine.max_chip_x, machine.max_chip_y))
    if print_all_chips:
        print('machine chips: {}'.format(list(machine.chips)))

    all_cores = []
    for chip in machine.chips:
        all_cores.append(CoreSubset(chip.x, chip.y, range(1, 17)))

    all_cores = CoreSubsets(core_subsets=all_cores)

    cores_finished = txrx.get_cores_in_state(all_cores, CPUState.FINISHED)
    cores_running = txrx.get_cores_in_state(all_cores, CPUState.RUNNING)
    cores_watchdog = txrx.get_cores_in_state(all_cores, CPUState.WATCHDOG)

    for (x, y, p), _ in cores_running:
        if p not in IGNORED_IDS:
            print('run core: {} {} {}'.format(x, y, p))

    for (x, y, p), _ in cores_finished:
        print('finished core: {} {} {}'.format(x, y, p))

    for (x, y, p), _ in cores_watchdog:
        print('watchdog core: {} {} {}'.format(x, y, p))


def _make_transceiver(config, host=None):
    """
    :param BoardTestConfiguration config:
    :param str host:
    :rtype: Transceiver
    """
    config.set_up_remote_board()
    if host is None:
        host = config.remotehost

    print("talking to SpiNNaker system at {}".format(host))
    return create_transceiver_from_hostname(
        host, config.board_version,
        ignore_cores=CoreSubsets(),
        ignore_chips=CoreSubsets(core_subsets=[]),
        bmp_connection_data=config.bmp_names,
        auto_detect_bmp=config.auto_detect_bmp)


def main():
    """ Runs the script.
    """
    ap = argparse.ArgumentParser(
        description="Check the state of a SpiNNaker machine.")
    ap.add_argument(
        "-a", "--appid", help="the application ID to check", type=int,
        default=17)
    ap.add_argument(
        "-n", "--noprintchips", action="store_true", default=False,
        help=("don't print all the chips out; avoids a great deal of "
              "output for large machines"))
    ap.add_argument(
        "host", default=None, nargs='?',
        help="the hostname or IP address of the SpiNNaker machine to inspect")
    args = ap.parse_args()
    # These ought to be parsed from command line arguments
    app_id = args.appid
    print_chips = not args.noprintchips

    try:
        from board_test_configuration import BoardTestConfiguration
        config = BoardTestConfiguration()
    except ImportError:
        print("cannot read board test configuration")
        sys.exit(1)
    transceiver = _make_transceiver(config, args.host)
    try:
        get_cores_in_run_state(transceiver, app_id, print_chips)
    finally:
        transceiver.close()


if __name__ == "__main__":  # pragma: no cover
    main()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
from spinnman.messages.scp.impl import ReadLink, ReadMemory
from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from spinnman.constants import UDP_MESSAGE_MAX_SIZE


class ReadMemoryProcess(AbstractMultiConnectionProcess):
    """ A process for reading memory on a SpiNNaker chip.
    """
    __slots__ = [
        "_view"]

    def __init__(self, connection_selector):
        """
        :param connection_selector:
        :type connection_selector:
            AbstractMultiConnectionProcessConnectionSelector
        """
        super().__init__(connection_selector)
        self._view = None

    def __handle_response(self, offset, response):
        self._view[offset:offset + response.length] = response.data[
            response.offset:response.offset + response.length]

    def read_memory(self, x, y, p, base_address, length):
        """
        :param int x:
        :param int y:
        :param int p:
        :param int base_address:
        :param int length:
        :rtype: bytearray
        """
        return self._read_memory(
            base_address, length,
            functools.partial(ReadMemory, x=x, y=y, cpu=p))

    def read_link_memory(self, x, y, p, link, base_address, length):
        """
        :param int x:
        :param int y:
        :param int p:
        :param int link:
        :param int base_address:
        :param int length:
        :rtype: bytearray
        """
        return self._read_memory(
            base_address, length,
            functools.partial(ReadLink, x=x, y=y, cpu=p, link=link))

    def _read_memory(self, base_address, length, packet_class):
        data = bytearray(length)
        self._view = memoryview(data)
        n_bytes = length
        offset = 0
        while n_bytes > 0:
            bytes_to_get = min((n_bytes, UDP_MESSAGE_MAX_SIZE))
            self._send_request(
                packet_class(
                    base_address=base_address + offset, size=bytes_to_get),
                functools.partial(self.__handle_response, offset))
            n_bytes -= bytes_to_get
            offset += bytes_to_get

        self._finish()
        self.check_for_error()

        return data



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .udp_connection import UDPConnection
from spinnman.constants import UDP_BOOT_CONNECTION_DEFAULT_PORT

_BOOTROM_SPINN_PORT = 54321  # Matches SPINN_PORT in spinnaker_bootROM


class IPAddressesConnection(UDPConnection):
    """ A connection that detects any UDP packet that is transmitted by\
        SpiNNaker boards prior to boot
    """
    __slots__ = []

    def __init__(self, local_host=None,
                 local_port=UDP_BOOT_CONNECTION_DEFAULT_PORT):
        super().__init__(local_host=local_host, local_port=local_port)

    def supports_sends_message(self, message):  # @UnusedVariable
        # pylint: disable=unused-argument
        return False

    def receive_ip_address(self, timeout=None):
        try:
            (_, (ip_address, port)) = self.receive_with_address(timeout)
            if port == _BOOTROM_SPINN_PORT:
                return ip_address
        except Exception:  # pylint: disable=broad-except
            pass
        return None

    def __repr__(self):
        return "IPAddressesConnection(local_host={}, local_port={})".format(
            self.local_ip_address, self.local_port)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
from spinnman.exceptions import (
    SpinnmanInvalidPacketException, SpinnmanInvalidParameterTypeException)
from .eieio_command_message import EIEIOCommandMessage
from .eieio_command_header import EIEIOCommandHeader
from spinnman.constants import EIEIO_COMMAND_IDS

_PATTERN_BBBB = struct.Struct("<BBBB")
_PATTERN_BBII = struct.Struct("<BBII")
_PATTERN_xxBBII = struct.Struct("<xxBBII")
_PATTERN_BB = struct.Struct("<BB")
_PATTERN_BBBBII = struct.Struct("<BBBBII")


class SpinnakerRequestReadData(EIEIOCommandMessage):
    """ Message used in the context of the buffering output mechanism which is\
        sent from the SpiNNaker system to the host computer to signal that\
        some data is available to be read
    """
    __slots__ = [
        "_header",
        "_requests"]

    def __init__(self, x, y, p, region_id, sequence_no, n_requests,
                 channel, start_address, space_to_be_read):
        # pylint: disable=too-many-arguments
        if not isinstance(channel, list):
            channel = [channel]

        if not isinstance(region_id, list):
            region_id = [region_id]

        if not isinstance(start_address, list):
            start_address = [start_address]

        if not isinstance(space_to_be_read, list):
            space_to_be_read = [space_to_be_read]

        if len(start_address) != n_requests or \
                len(space_to_be_read) != n_requests or \
                len(channel) != n_requests or \
                len(region_id) != n_requests:
            raise SpinnmanInvalidPacketException(
                "SpinnakerRequestReadData",
                "The format for a SpinnakerRequestReadData packet is "
                "invalid: {0:d} request(s), {1:d} start address(es) "
                "defined, {2:d} space(s) to be read defined, {3:d} region(s) "
                "defined, {4:d} channel(s) defined".format(
                    n_requests, len(start_address), len(space_to_be_read),
                    len(region_id), len(channel)))

        super().__init__(EIEIOCommandHeader(
            EIEIO_COMMAND_IDS.SPINNAKER_REQUEST_READ_DATA))
        self._header = _SpinnakerRequestReadDataHeader(
            x, y, p, n_requests, sequence_no)
        self._requests = _SpinnakerRequestReadDataRequest(
            channel, region_id, start_address, space_to_be_read)

    @property
    def x(self):
        return self._header.x

    @property
    def y(self):
        return self._header.y

    @property
    def p(self):
        return self._header.p

    @property
    def n_requests(self):
        return self._header.n_requests

    @property
    def sequence_no(self):
        return self._header.sequence_no

    def channel(self, request_id):
        return self._requests.channel(request_id)

    def region_id(self, request_id):
        return self._requests.region_id(request_id)

    def start_address(self, request_id):
        return self._requests.start_address(request_id)

    def space_to_be_read(self, request_id):
        return self._requests.space_to_be_read(request_id)

    @staticmethod
    def get_min_packet_length():
        return 16

    @staticmethod
    def from_bytestring(command_header, data, offset):
        (y, x, processor_and_requests, sequence_no) = \
            _PATTERN_BBBB.unpack_from(data, offset)
        p = (processor_and_requests >> 3) & 0x1F
        n_requests = processor_and_requests & 0x7
        offset += 4
        channel = list()
        region_id = list()
        start_address = list()
        space_to_be_read = list()

        for i in range(n_requests):
            if i == 0:
                request_channel, request_region_id, request_start_address, \
                    request_space_to_be_read = _PATTERN_BBII.unpack_from(
                        data, offset)
                offset += 10
            else:
                request_channel, request_region_id, request_start_address, \
                    request_space_to_be_read = _PATTERN_xxBBII.unpack_from(
                        data, offset)
                offset += 12
            channel.append(request_channel)
            region_id.append(request_region_id)
            start_address.append(request_start_address)
            space_to_be_read.append(request_space_to_be_read)
        return SpinnakerRequestReadData(
            x, y, p, region_id, sequence_no, n_requests, channel,
            start_address, space_to_be_read)

    @property
    def bytestring(self):
        byte_string = super().bytestring
        byte_string += _PATTERN_BB.pack(self.x, self.y)
        n_requests = self.n_requests
        processor_and_request = (self.p << 3) | n_requests

        for i in range(n_requests):
            if i == 0:
                byte_string += _PATTERN_BBBBII.pack(
                    processor_and_request, self.sequence_no,
                    self.channel(i), self.region_id(i),
                    self.start_address(0), self.space_to_be_read(i))
            else:
                byte_string += _PATTERN_xxBBII.pack(
                    self.channel(i), self.region_id(i),
                    self.start_address(0), self.space_to_be_read(i))
        return byte_string


class _SpinnakerRequestReadDataHeader(object):
    """ Contains the position of the core in the machine (x, y, p), the number\
        of requests and a sequence number
    """
    __slots__ = [
        "_n_requests",
        "_sequence_no",
        "_p", "_x", "_y"]

    def __init__(self, x, y, p, n_requests, sequence_no):
        # pylint: disable=too-many-arguments
        self._x = x
        self._y = y
        self._p = p
        self._n_requests = n_requests
        self._sequence_no = sequence_no

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def p(self):
        return self._p

    @property
    def sequence_no(self):
        return self._sequence_no

    @property
    def n_requests(self):
        return self._n_requests


class _SpinnakerRequestReadDataRequest(object):
    """ Contains a set of requests which refer to the channels used
    """
    __slots__ = [
        "_channel",
        "_region_id",
        "_start_address",
        "_space_to_be_read"]

    def __init__(self, channel, region_id, start_address, space_to_be_read):
        if not isinstance(channel, list):
            self._channel = [channel]
        else:
            self._channel = channel

        if not isinstance(region_id, list):
            self._region_id = [region_id]
        else:
            self._region_id = region_id

        if not isinstance(start_address, list):
            self._start_address = [start_address]
        else:
            self._start_address = start_address

        if not isinstance(space_to_be_read, list):
            self._space_to_be_read = [space_to_be_read]
        else:
            self._space_to_be_read = space_to_be_read

    def channel(self, request_id):
        if len(self._channel) > request_id:
            return self._channel[request_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer", "channel request needs to be"
            "comprised between 0 and {0:d}; current value: {1:d}".format(
                len(self._channel) - 1, request_id))

    def region_id(self, request_id):
        if len(self._region_id) > request_id:
            return self._region_id[request_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer", "region ID request needs to be"
            "comprised between 0 and {0:d}; current value: {1:d}".format(
                len(self._region_id) - 1, request_id))

    def start_address(self, request_id):
        if len(self._start_address) > request_id:
            return self._start_address[request_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer", "start address request needs to be"
            "comprised between 0 and {0:d}; current value: {1:d}".format(
                len(self._start_address) - 1, request_id))

    def space_to_be_read(self, request_id):
        if len(self._space_to_be_read) > request_id:
            return self._space_to_be_read[request_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer", "space to be read request needs to be"
            "comprised between 0 and {0:d}; current value: {1:d}".format(
                len(self._space_to_be_read) - 1, request_id))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinnman.exceptions import SpinnmanInvalidParameterException
from spinnman.constants import ROUTER_REGISTER_REGISTERS
from spinnman.model.enums.router_error import RouterError


class RouterDiagnostics(object):
    """ Represents a set of diagnostic information available from a chip\
        router.
    """
    __slots__ = [
        "_error_status",
        "_mon",
        "_register_values",
        "_wait_1",
        "_wait_2"]

    def __init__(self, control_register, error_status, register_values):
        """
        :param int control_register: The value of the control register
        :param int error_status: The value of the error_status
        :param list(int) register_values:
            The values of the 16 router registers
        :raise SpinnmanInvalidParameterException:
            If the number of register values is not 16
        """
        if len(register_values) != 16:
            raise SpinnmanInvalidParameterException(
                "len(register_values)",
                str(len(ROUTER_REGISTER_REGISTERS)),
                "There must be exactly 16 register values")

        self._mon = (control_register >> 8) & 0x1F
        self._wait_1 = (control_register >> 16) & 0xFF
        self._wait_2 = (control_register >> 24) & 0xFF

        self._error_status = error_status

        self._register_values = register_values

    @property
    def mon(self):
        """ The "mon" part of the control register

        :rtype: int
        """
        return self._mon

    @property
    def wait_1(self):
        """ The "wait_1" part of the control register

        :rtype: int
        """
        return self._wait_1

    @property
    def wait_2(self):
        """ The "wait_2" part of the control register

        :rtype: int
        """
        return self._wait_2

    @property
    def error_status(self):
        """ The error status

        :rtype: int
        """
        return self._error_status

    @property
    def error_count(self):
        """ The count of errors

        :rtype: int
        """
        return self._error_status & 0xFF

    @property
    def errors_set(self):
        """ A list of errors that have been detected

        :rtype: list(RouterError)
        """
        return [
            error for error in RouterError if error.value & self._error_status]

    @property
    def n_local_multicast_packets(self):
        """ The number of multicast packets received from local cores.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.LOC_MC.value]

    @property
    def n_external_multicast_packets(self):
        """ The number of multicast packets received from external links.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.EXT_MC.value]

    @property
    def n_dropped_multicast_packets(self):
        """ The number of multicast packets received that were dropped.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.DUMP_MC.value]

    @property
    def n_local_peer_to_peer_packets(self):
        """ The number of peer-to-peer packets received from local cores.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.LOC_PP.value]

    @property
    def n_external_peer_to_peer_packets(self):
        """ The number of peer-to-peer packets received from external links.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.EXT_PP.value]

    @property
    def n_dropped_peer_to_peer_packets(self):
        """ The number of peer-to-peer packets received that were dropped.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.DUMP_PP.value]

    @property
    def n_local_nearest_neighbour_packets(self):
        """ The number of nearest-neighbour packets received from local cores.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.LOC_NN.value]

    @property
    def n_external_nearest_neighbour_packets(self):
        """ The number of nearest-neighbour packets received from external\
            links.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.EXT_NN.value]

    @property
    def n_dropped_nearest_neighbour_packets(self):
        """ The number of nearest-neighbour packets received that were dropped.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.DUMP_NN.value]

    @property
    def n_local_fixed_route_packets(self):
        """ The number of fixed-route packets received from local cores.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.LOC_FR.value]

    @property
    def n_external_fixed_route_packets(self):
        """ The number of fixed-route packets received from external links.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.EXT_FR.value]

    @property
    def n_dropped_fixed_route_packets(self):
        """ The number of fixed-route packets received that were dropped.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.DUMP_FR.value]

    @property
    def user_0(self):
        """ The number of packets counted by the user 0 router diagnostic\
            filter.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.USER_0.value]

    @property
    def user_1(self):
        """ The number of packets counted by the user 1 router diagnostic\
            filter.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.USER_1.value]

    @property
    def user_2(self):
        """ The number of packets counted by the user 2 router diagnostic\
            filter.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.USER_2.value]

    @property
    def user_3(self):
        """ The number of packets counted by the user 3 router diagnostic\
            filter.

        :rtype: int
        """
        return self._register_values[ROUTER_REGISTER_REGISTERS.USER_3.value]

    @property
    def user_registers(self):
        """ The values in the user control registers.

        :return: An array of 4 values
        :rtype: list(int)
        """
        return self._register_values[
            ROUTER_REGISTER_REGISTERS.USER_0.value:
            ROUTER_REGISTER_REGISTERS.USER_3.value + 1]

    @property
    def registers(self):
        """ The values in all of the registers.  Can be used to directly\
            access the registers if they have been programmed to give\
            different values

        :return: An array of 16 values
        :rtype: array(int)
        """
        return self._register_values



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from .connection import Connection


class MulticastSender(Connection, metaclass=AbstractBase):
    """ A sender of Multicast messages
    """

    __slots__ = ()

    @abstractmethod
    def get_input_chips(self):
        """ Get a list of chips which identify the chips to which this sender\
            can send multicast packets directly

        :return: The coordinates, (x, y), of the chips
        :rtype: iterable(tuple(int, int))
        """

    @abstractmethod
    def send_multicast_message(self, multicast_message):
        """ Sends a SpiNNaker multicast message using this connection

        :param MulticastMessage multicast_message: The message to be sent
        :raise SpinnmanIOException:
            If there is an error sending the message
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import struct
from collections import defaultdict, OrderedDict
from spinnman.model import IOBuffer
from spinnman.utilities.utility_functions import get_vcpu_address
from spinnman.messages.scp.impl import ReadMemory
from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from spinnman.constants import UDP_MESSAGE_MAX_SIZE, CPU_IOBUF_ADDRESS_OFFSET

_ENCODING = "ascii"
_ONE_WORD = struct.Struct("<I")
_FIRST_IOBUF = struct.Struct("<I8xI")


class ReadIOBufProcess(AbstractMultiConnectionProcess):
    """ A process for reading IOBUF memory (mostly log messages) from a\
        SpiNNaker core.
    """
    __slots__ = [
        "_extra_reads",
        "_iobuf",
        "_iobuf_address",
        "_iobuf_view",
        "_next_reads"]

    def __init__(self, connection_selector):
        """
        :param connection_selector:
        :type connection_selector:
            AbstractMultiConnectionProcessConnectionSelector
        """
        super().__init__(connection_selector)

        # A dictionary of (x, y, p) -> iobuf address
        self._iobuf_address = dict()

        # A dictionary of (x, y, p) -> OrderedDict(n) -> bytearray
        self._iobuf = defaultdict(OrderedDict)

        # A dictionary of (x, y, p) -> OrderedDict(n) -> memoryview
        self._iobuf_view = defaultdict(OrderedDict)

        # A list of extra reads that need to be done as a result of the first
        # read = list of (x, y, p, n, base_address, size, offset)
        self._extra_reads = list()

        # A list of next reads that need to be done as a result of the first
        # read = list of (x, y, p, n, next_address, first_read_size)
        self._next_reads = list()

    def _request_iobuf_address(self, iobuf_size, x, y, p):
        base_address = get_vcpu_address(p) + CPU_IOBUF_ADDRESS_OFFSET
        self._send_request(
            ReadMemory(x, y, base_address, 4),
            functools.partial(self._handle_iobuf_address_response,
                              iobuf_size, x, y, p))

    def _handle_iobuf_address_response(
            self, iobuf_size, x, y, p, response):
        iobuf_address, = _ONE_WORD.unpack_from(response.data, response.offset)
        if iobuf_address != 0:
            first_read_size = min((iobuf_size + 16, UDP_MESSAGE_MAX_SIZE))
            self._next_reads.append((
                x, y, p, 0, iobuf_address, first_read_size))

    def _request_iobuf_region_tail(self, extra_region):
        (x, y, p, n, base_address, size, offset) = extra_region
        self._send_request(
            ReadMemory(x, y, base_address, size),
            functools.partial(self._handle_extra_iobuf_response,
                              x, y, p, n, offset))

    def _handle_extra_iobuf_response(
            self, x, y, p, n, offset, response):
        view = self._iobuf_view[x, y, p][n]
        view[offset:offset + response.length] = response.data[
            response.offset:response.offset + response.length]

    def _request_iobuf_region(self, region):
        (x, y, p, n, next_address, first_read_size) = region
        self._send_request(
            ReadMemory(x, y, next_address, first_read_size),
            functools.partial(self._handle_first_iobuf_response,
                              x, y, p, n, next_address, first_read_size))

    def _handle_first_iobuf_response(self, x, y, p, n, base_address,
                                     first_read_size, response):
        # pylint: disable=too-many-arguments

        # Unpack the iobuf header
        (next_address, bytes_to_read) = _FIRST_IOBUF.unpack_from(
            response.data, response.offset)

        # Create a buffer for the data
        data = bytearray(bytes_to_read)
        view = memoryview(data)
        self._iobuf[x, y, p][n] = data
        self._iobuf_view[x, y, p][n] = view

        # Put the data from this packet into the buffer
        packet_bytes = response.length - 16
        if packet_bytes > bytes_to_read:
            packet_bytes = bytes_to_read
        if packet_bytes > 0:
            offset = response.offset + 16
            view[0:packet_bytes] = response.data[
                offset:(offset + packet_bytes)]

        bytes_to_read -= packet_bytes
        base_address += packet_bytes + 16
        read_offset = packet_bytes

        # While more reads need to be done to read the data
        while bytes_to_read > 0:
            # Read the next bit of memory making up the buffer
            next_bytes_to_read = min((bytes_to_read, UDP_MESSAGE_MAX_SIZE))
            self._extra_reads.append(
                (x, y, p, n, base_address, next_bytes_to_read, read_offset))
            base_address += next_bytes_to_read
            read_offset += next_bytes_to_read
            bytes_to_read -= next_bytes_to_read

        # If there is another IOBuf buffer, read this next
        if next_address != 0:
            self._next_reads.append(
                (x, y, p, n + 1, next_address, first_read_size))

    def read_iobuf(self, iobuf_size, core_subsets):
        """
        :param int iobuf_size:
        :param ~spinn_machine.CoreSubsets core_subsets:
        :rtype: iterable(IOBuffer)
        """
        # Get the iobuf address for each core
        for core_subset in core_subsets:
            x = core_subset.x
            y = core_subset.y
            for p in core_subset.processor_ids:
                self._request_iobuf_address(iobuf_size, x, y, p)
        self._finish()
        self.check_for_error()

        # Run rounds of the process until reading is complete
        while self._extra_reads or self._next_reads:
            # Process the extra iobuf reads needed
            while self._extra_reads:
                self._request_iobuf_region_tail(self._extra_reads.pop())

            # Process the next iobuf reads needed
            while self._next_reads:
                self._request_iobuf_region(self._next_reads.pop())

            # Finish this round
            self._finish()
            self.check_for_error()

        for core_subset in core_subsets:
            x = core_subset.x
            y = core_subset.y
            for p in core_subset.processor_ids:
                iobuf = ""
                for item in self._iobuf[x, y, p].values():
                    iobuf += item.decode(_ENCODING)
                yield IOBuffer(x, y, p, iobuf)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict
import logging
import functools
import struct
from os.path import join
from spinn_utilities.log import FormatAdapter
from spinn_machine import (
    Router, Chip, SDRAM, Link, machine_from_size)
from spinn_machine import Machine
from spinn_machine.machine_factory import machine_repair
from spinnman.constants import (
    ROUTER_REGISTER_P2P_ADDRESS, SYSTEM_VARIABLE_BASE_ADDRESS)
from spinnman.messages.spinnaker_boot import (
    SystemVariableDefinition)
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException
from spinnman.messages.scp.impl import ReadMemory, ReadLink, GetChipInfo
from spinnman.model import P2PTable
from spinnman.model.enums import CPUState
from .abstract_multi_connection_process import AbstractMultiConnectionProcess

logger = FormatAdapter(logging.getLogger(__name__))

REPORT_FILE = "Ignores_report.rpt"


class GetMachineProcess(AbstractMultiConnectionProcess):
    """ A process for getting the machine details over a set of connections.
    """
    __slots__ = [
        "_chip_info",
        # Used if there are any ignores with ip addresses
        # Holds a mapping from ip to board root (x,y)
        "_ethernets",
        "_ignore_chips",
        "_ignore_cores",
        # Holds a map from x,y to a set of virtual cores to ignores
        "_ignore_cores_map",
        "_ignore_links",
        "_max_sdram_size",
        "_p2p_column_data",
        # Used if there are any ignore core requests
        # Holds a mapping from (x,y) to a mapping of phsyical to virtual core
        "_virtual_map",
        # Directory to put the ingore report if required
        "_default_report_directory",
        # Ignore report file path for ignre report.
        # Kept as None until first write
        "_report_file"]

    def __init__(self, connection_selector, ignore_chips, ignore_cores,
                 ignore_links, max_sdram_size=None,
                 default_report_directory=None):
        """
        :param connection_selector:
        :type connection_selector:
            AbstractMultiConnectionProcessConnectionSelector
        """
        # pylint: disable=too-many-arguments
        super().__init__(connection_selector)

        self._ignore_chips = ignore_chips if ignore_chips is not None else {}
        self._ignore_cores = ignore_cores if ignore_cores is not None else {}
        self._ignore_cores_map = defaultdict(set)
        self._ignore_links = ignore_links if ignore_links is not None else {}

        self._max_sdram_size = max_sdram_size

        self._p2p_column_data = list()

        # A dictionary of (x, y) -> ChipInfo
        self._chip_info = dict()

        # Set ethernets to None meaning not computed yet
        self._ethernets = None
        self._virtual_map = {}
        self._default_report_directory = default_report_directory
        self._report_file = None

    def _make_chip(self, chip_info, machine):
        """ Creates a chip from a ChipSummaryInfo structure.

        :param ChipSummaryInfo chip_info:
            The ChipSummaryInfo structure to create the chip from
        :return: The created chip
        :rtype: ~spinn_machine.Chip
        """
        # Create the down cores set if any
        n_cores = min(chip_info.n_cores, Machine.max_cores_per_chip())
        core_states = chip_info.core_states
        down_cores = self._ignore_cores_map.get(
            (chip_info.x, chip_info.y), None)
        for i in range(1, n_cores):
            if core_states[i] != CPUState.IDLE:
                self._report_ignore(
                    "Not using core {}, {}, {} in state {}",
                    chip_info.x, chip_info.y, i,  core_states[i])
                if down_cores is None:
                    down_cores = set()
                down_cores.add(i)

        # Create the router
        router = self._make_router(chip_info, machine)

        # Create the chip's SDRAM object
        sdram_size = chip_info.largest_free_sdram_block
        if (self._max_sdram_size is not None and
                sdram_size > self._max_sdram_size):
            sdram_size = self._max_sdram_size
        sdram = SDRAM(size=sdram_size)

        # Create the chip
        return Chip(
            x=chip_info.x, y=chip_info.y, n_processors=n_cores,
            router=router, sdram=sdram,
            ip_address=chip_info.ethernet_ip_address,
            nearest_ethernet_x=chip_info.nearest_ethernet_x,
            nearest_ethernet_y=chip_info.nearest_ethernet_y,
            down_cores=down_cores, parent_link=chip_info.parent_link)

    def _make_router(self, chip_info, machine):
        """
        :param ChipSummaryInfo chip_info:
        :param ~spinn_machine.Machine machine:
        :rtype: ~spinn_machine.Router
        """
        links = list()
        for link in chip_info.working_links:
            dest_xy = machine.xy_over_link(chip_info.x, chip_info.y, link)
            if dest_xy in self._chip_info:
                links.append(Link(
                    chip_info.x, chip_info.y, link, dest_xy[0], dest_xy[1]))
            else:
                logger.warning(
                    "Link {},{}:{} points to {} but that is not included ",
                    chip_info.x, chip_info.y, link, dest_xy)

        return Router(
            links=links,
            n_available_multicast_entries=(
                chip_info.n_free_multicast_routing_entries))

    def __receive_p2p_data(self, column, scp_read_response):
        """
        :param int column:
        :param _SCPReadMemoryResponse scp_read_response:
        """
        self._p2p_column_data[column] = (
            scp_read_response.data, scp_read_response.offset)

    def _receive_chip_info(self, scp_read_chip_info_response):
        """
        :param GetChipInfoResponse scp_read_chip_info_response:
        """
        chip_info = scp_read_chip_info_response.chip_info
        self._chip_info[chip_info.x, chip_info.y] = chip_info

    def _receive_error(self, request, exception, tb, connection):
        """
        :param AbstractSCPRequest request:
        :param Exception exception:
        """
        # If we get an ReadLink with a
        # SpinnmanUnexpectedResponseCodeException, this is a failed link
        # and so can be ignored
        if isinstance(request, ReadLink):
            if isinstance(exception, SpinnmanUnexpectedResponseCodeException):
                return
        super()._receive_error(request, exception, tb, connection)

    def get_machine_details(self, boot_x, boot_y, width, height,
                            repair_machine, ignore_bad_ethernets):
        """
        :param int boot_x:
        :param int boot_y:
        :param int width:
        :param int height:
        :param bool repair_machine:
        :param bool ignore_bad_ethernets:
        :rtype: ~spinn_machine.Machine
        """
        # Get the P2P table - 8 entries are packed into each 32-bit word
        p2p_column_bytes = P2PTable.get_n_column_bytes(height)
        self._p2p_column_data = [None] * width
        for column in range(width):
            offset = P2PTable.get_column_offset(column)
            self._send_request(
                ReadMemory(
                    x=boot_x, y=boot_y,
                    base_address=(ROUTER_REGISTER_P2P_ADDRESS + offset),
                    size=p2p_column_bytes),
                functools.partial(self.__receive_p2p_data, column))
        self._finish()
        self.check_for_error()
        p2p_table = P2PTable(width, height, self._p2p_column_data)

        # Get the chip information for each chip
        for (x, y) in p2p_table.iterchips():
            self._send_request(GetChipInfo(x, y), self._receive_chip_info)
        self._finish()
        try:
            self.check_for_error()
        except Exception:  # pylint: disable=broad-except
            # Ignore errors so far, as any error here just means that a chip
            # is down that wasn't marked as down
            pass

        # Warn about unexpected missing chips
        for (x, y) in p2p_table.iterchips():
            if (x, y) not in self._chip_info:
                logger.warning(
                    "Chip {}, {} was expected but didn't reply", x, y)

        machine = machine_from_size(width, height)
        self._preprocess_ignore_chips(machine)
        self._process_ignore_links(machine)
        self._preprocess_ignore_cores(machine)

        return self._fill_machine(
            machine, repair_machine, ignore_bad_ethernets)

    def _fill_machine(
            self, machine, repair_machine, ignore_bad_ethernets):
        """
        :param ~spinn_machine.Machine machine:
        :param bool repair_machine:
        :param bool ignore_bad_ethernets:
        :rtype: ~spinn_machine.Machine
        """
        for chip_info in sorted(
                self._chip_info.values(), key=lambda chip: (chip.x, chip.y)):
            if (chip_info.ethernet_ip_address is not None and
                    (chip_info.x != chip_info.nearest_ethernet_x
                     or chip_info.y != chip_info.nearest_ethernet_y)):
                if ignore_bad_ethernets:
                    logger.warning(
                        "Chip {}:{} claimed it has ip address: {}. "
                        "This ip will not be used.",
                        chip_info.x, chip_info.y,
                        chip_info.ethernet_ip_address)
                    chip_info.ethernet_ip_address = None
                else:
                    logger.warning(
                        "Not using chip {}:{} as it has an unexpected "
                        "ip address: {}", chip_info.x, chip_info.y,
                        chip_info.ethernet_ip_address)
                    continue

            # If the above has not continued, add the chip
            machine.add_chip(self._make_chip(chip_info, machine))

        machine.validate()
        return machine_repair(machine, repair_machine)

    # Stuff below here is purely for dealing with ignores

    def _process_ignore_links(self, machine):
        """
        Processes the collection of ignore links to remove then from chipinfo

        Converts any local x, y, IP address to global xy

        Discards any ignores which are known to have no\
        affect based on the already read chip_info

        Also removes any inverse links

        Logs all actions except for ignores with unused IP addresses

        :param ~spinn_machine.Machine machine:
            An empty machine to handle wraparounds
        """
        if len(self._ignore_links) == 0:
            return

        for ignore in self._ignore_links:
            global_xy = self._ignores_local_to_global(
                ignore.x, ignore.y, ignore.ip_address, machine)
            if global_xy is None:
                continue
            chip_info = self._chip_info.get(global_xy, None)
            if chip_info is None:
                self._report_ignore(
                    "Discarding ignore link on chip {} as it is not/ no longer"
                    " in info", global_xy)
                continue
            link = ignore.link
            if link in chip_info.working_links:
                chip_info.working_links.remove(link)
                self._report_ignore(
                    "On chip {} ignoring link:{}", global_xy, link)
                # ignore the inverse link too
                inv_xy = machine.xy_over_link(global_xy[0], global_xy[1], link)
                if inv_xy in self._chip_info:
                    inv_chip_info = self._chip_info[inv_xy]
                    inv_link = (link + 3) % 6
                    if inv_link in inv_chip_info.working_links:
                        inv_chip_info.working_links.remove(inv_link)
                        self._report_ignore(
                            "On chip {} ignoring link {} as it is the inverse "
                            "of link {} on chip {}", inv_xy, inv_link, link,
                            global_xy)
            else:
                self._report_ignore(
                    "Discarding ignore link {} on chip {} as it is not/"
                    "no longer in info", link, global_xy)

    def _preprocess_ignore_cores(self, machine):
        """
        Converts the collection of ignore cores into a map of ignore by xy

        Converts any local x, y ipaddress to global xy

        Discards (with log messages) any ignores which are known to have no\
        affect based on the already read chip_info

        Converts any physical  cores to virtual ones.\
        Core numbers <= 0 are assumed to be 0 - physical_id

        :param ~spinn_machine.Machine machine:
            An empty machine to handle wraparounds
        """
        if len(self._ignore_cores) == 0:
            return
        # Convert by ip to global
        for ignore in self._ignore_cores:
            global_xy = self._ignores_local_to_global(
                ignore.x, ignore.y, ignore.ip_address, machine)
            if global_xy is None:
                continue
            p = self._get_virtual_p(global_xy, ignore.p)
            if p is not None:
                self._ignore_cores_map[global_xy].add(p)

    def _preprocess_ignore_chips(self, machine):
        """
        Processes the collection of ignore chips and discards their chipinfo

        Converts any local x, y ipaddress to global xy

        Discards any ignores which are known to have no\
        affect based on the already read chip_info

        Logs all actions except for ignores with unused ip addresses

        :param ~spinn_machine.Machine machine:
            An empty machine to handle wraparounds
        """
        for ignore in self._ignore_chips:
            # Convert by ip to global
            global_xy = self._ignores_local_to_global(
                ignore.x, ignore.y, ignore.ip_address, machine)
            if global_xy is None:
                continue  # Never on this machine
            chip_info = self._chip_info.pop(global_xy, None)
            if chip_info is None:
                continue  # Already ignored maybe by a dead chip list
            self._report_ignore("Chip {} will be ignored", global_xy)
            for link in chip_info.working_links:
                # ignore the inverse link
                inv_xy = machine.xy_over_link(global_xy[0], global_xy[1], link)
                if inv_xy in self._chip_info:
                    inv_chip_info = self._chip_info[inv_xy]
                    inv_link = (link + 3) % 6
                    if inv_link in inv_chip_info.working_links:
                        inv_chip_info.working_links.remove(inv_link)
                        self._report_ignore(
                            "On chip {} ignoring link {} as it points to "
                            "ignored chip chip {}",
                            inv_xy, inv_link, global_xy)

    def _ignores_local_to_global(self, local_x, local_y, ip_address, machine):
        """
        :param int local_x:
        :param int local_y:
        :param str ip_address:
        :param ~spinn_machine.Machine machine:
        :rtype: tuple(int,int)
        """
        if ip_address is None:
            global_xy = (local_x, local_y)
        else:
            ethernet = self._ethernet_by_ipaddress(ip_address)
            if ethernet is None:
                self._report_ignore(
                    "Ignore with ip:{} will be discarded as no board with "
                    "that ip in this machine", ip_address)
                return None
            global_xy = machine.get_global_xy(
                local_x, local_y, ethernet[0], ethernet[1])
            self._report_ignore(
                "Ignores for local x:{} y:{} and ip:{} map to global {} with "
                "root {}", local_x, local_y, ip_address, global_xy,
                self._chip_info[(0, 0)].ethernet_ip_address)

        if global_xy in self._chip_info:
            return global_xy
        else:
            self._report_ignore(
                "Ignore for global chip {} will be discarded as no such chip "
                "in this machine", global_xy)
            return None

    def _ethernet_by_ipaddress(self, ip_address):
        """
        :param str ip_address:
        :rtype: tuple(int,int)
        """
        if self._ethernets is None:
            self._ethernets = dict()
            for chip_info in self._chip_info.values():
                if chip_info.ethernet_ip_address is not None:
                    self._ethernets[chip_info.ethernet_ip_address] = \
                        (chip_info.x, chip_info.y)
        return self._ethernets.get(ip_address, None)

    def _get_virtual_p(self, xy, p):
        """
        :param tuple(int,int) xy:
        :param int p:
        :rtype: int
        """
        if xy not in self._virtual_map:
            if xy not in self._chip_info:
                # Chip not part of board so ignore
                return None
            p_to_v = SystemVariableDefinition.physical_to_virtual_core_map
            ba = SYSTEM_VARIABLE_BASE_ADDRESS + p_to_v.offset
            self._send_request(
                ReadMemory(
                    x=xy[0], y=xy[1],
                    base_address=ba,
                    size=p_to_v.array_size),
                functools.partial(
                    self._receive_physical_to_virtual_core_map, xy))
            self._finish()

        if p > 0:
            self._report_ignore("On chip {} ignoring core {}", xy, p)
            return p
        else:
            virtual_map = self._virtual_map[xy]
            if 0-p not in virtual_map:
                self._report_ignore(
                    "On chip {} physical core {} was not used "
                    "so ignore is being discarded.".format(xy, -p))
                return None
            virtual_p = virtual_map[0-p]
            if virtual_p == 0:
                self._report_ignore(
                    "On chip {} physical core {} was used as the monitor "
                    "so will NOT be ignored".format(xy, -p))
                return None
            self._report_ignore(
                "On chip {} ignoring core {} as it maps to physical "
                "core {}", xy, 0-p, virtual_p)
            return virtual_p

    def _receive_physical_to_virtual_core_map(self, xy, scp_read_response):
        """
        :param tuple(int,int) xy:
        :param _SCPReadMemoryResponse scp_read_response:
        """
        chipinfo = self._chip_info[xy]
        ip_address = self._chip_info[(0, 0)].ethernet_ip_address

        self._report_ignore(
            "Received physical_to_virtual_core_map for chip {} on {}",
            xy, ip_address)
        p_to_v_map = {}

        for i in range(
                scp_read_response.offset,
                scp_read_response.offset + chipinfo.n_cores):
            p_to_v_map[i-scp_read_response.offset] = \
                struct.unpack_from("b", scp_read_response.data, i)[0]
        # report_ignore (like logger) formats so dict as a param
        self._report_ignore("{}", p_to_v_map)
        self._virtual_map[xy] = p_to_v_map

    def _verify_virtual_to_physical_core_map(self, xy):
        """ Add this method to _get_virtual_p to verify the mappings.

        :param tuple(int,int) xy:
        """
        v_to_p = SystemVariableDefinition.virtual_to_physical_core_map
        self._send_request(
            ReadMemory(
                x=xy[0], y=xy[1],
                base_address=SYSTEM_VARIABLE_BASE_ADDRESS + v_to_p.offset,
                size=v_to_p.array_size),
            functools.partial(
                self._receive_virtual_to_physical_core_map, xy))
        self._finish()

    def _receive_virtual_to_physical_core_map(self, xy, scp_read_response):
        """
        :param tuple(int,int) xy:
        :param _SCPReadMemoryResponse scp_read_response:
        """
        p_to_v_map = self._virtual_map[xy]
        chipinfo = self._chip_info[xy]
        for i in range(
                scp_read_response.offset,
                scp_read_response.offset + chipinfo.n_cores):
            assert p_to_v_map[int(scp_read_response.data[i])] == \
                    i-scp_read_response.offset
        self._report_ignore(
            "Virtual_to_physical_core_map checks for chip {}", xy)

    def _report_ignore(self, message, *args):
        """
        Writes the ignore message by either creating or appending the report

        The implementation choice to reopen the file every time is not the\
        fastest but is the cleanest and safest for code that in default\
        conditions is never run.

        :param str message:
        """
        full_message = message.format(*args) + "\n"
        if self._report_file is None:
            if self._default_report_directory is None:
                self._report_file = REPORT_FILE
            else:
                self._report_file = join(
                    self._default_report_directory, REPORT_FILE)

        with open(self._report_file, "a") as r_file:
            r_file.write(full_message)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import socket
import select
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinnman.exceptions import SpinnmanIOException, SpinnmanTimeoutException
from spinnman.connections.abstract_classes import Connection
from .utils import (
    bind_socket, connect_socket, get_socket, get_socket_address, ping,
    resolve_host, set_receive_buffer_size)

logger = FormatAdapter(logging.getLogger(__name__))
_RECEIVE_BUFFER_SIZE = 1048576
_PING_COUNT = 5
_REPR_TEMPLATE = "UDPConnection(local={}:{}, remote={}:{})"


class UDPConnection(Connection):
    __slots__ = [
        "_can_send",
        "_local_ip_address",
        "_local_port",
        "_remote_ip_address",
        "_remote_port",
        "_socket"]

    def __init__(self, local_host=None, local_port=None, remote_host=None,
                 remote_port=None):
        """
        :param str local_host: The local host name or IP address to bind to.
            If not specified defaults to bind to all interfaces, unless
            remote_host is specified, in which case binding is done to the
            IP address that will be used to send packets
        :param int local_port: The local port to bind to, between 1025 and
            65535. If not specified, defaults to a random unused local port
        :param str remote_host: The remote host name or IP address to send
            packets to. If not specified, the socket will be available for
            listening only, and will throw and exception if used for sending
        :param int remote_port: The remote port to send packets to.  If
            remote_host is None, this is ignored.  If remote_host is specified
            specified, this must also be specified for the connection to allow
            sending
        :raise SpinnmanIOException:
            If there is an error setting up the communication channel
        """

        self._socket = get_socket()
        set_receive_buffer_size(self._socket, _RECEIVE_BUFFER_SIZE)

        # Get the host and port to bind to locally
        local_bind_host = "" if local_host is None else local_host
        local_bind_port = 0 if local_port is None else local_port
        bind_socket(self._socket, local_bind_host, local_bind_port)

        # Mark the socket as non-sending, unless the remote host is
        # specified - send requests will then cause an exception
        self._can_send = False
        self._remote_ip_address = None
        self._remote_port = None

        # Get the host to connect to remotely
        if remote_host is not None and remote_port is not None:
            self._remote_port = remote_port
            self._remote_ip_address = resolve_host(remote_host)
            connect_socket(self._socket, self._remote_ip_address, remote_port)
            self._can_send = True

        # Get the details of where the socket is connected
        self._local_ip_address, self._local_port = \
            get_socket_address(self._socket)

        # Set a general timeout on the socket
        self._socket.settimeout(1.0)

    @overrides(Connection.is_connected)
    def is_connected(self):
        # If this is not a sending socket, it is not connected
        if not self._can_send:
            return False

        # check if machine is active and on the network
        for _ in range(_PING_COUNT):
            # Assume connected if ping works
            if ping(self._remote_ip_address).returncode == 0:
                return True

        # If the ping fails this number of times, the host cannot be contacted
        return False

    @property
    def local_ip_address(self):
        """ The local IP address to which the connection is bound.

        :return: The local IP address as a dotted string, e.g., 0.0.0.0
        :rtype: str
        """
        return self._local_ip_address

    @property
    def local_port(self):
        """ The local port to which the connection is bound.

        :return: The local port number
        :rtype: int
        """
        return self._local_port

    @property
    def remote_ip_address(self):
        """ The remote IP address to which the connection is connected.

        :return: The remote IP address as a dotted string, or None if not
            connected remotely
        :rtype: str
        """
        return self._remote_ip_address

    @property
    def remote_port(self):
        """ The remote port to which the connection is connected.

        :return: The remote port, or None if not connected remotely
        :rtype: int
        """
        return self._remote_port

    def receive(self, timeout=None):
        """ Receive data from the connection

        :param float timeout: The timeout in seconds, or None to wait forever
        :return: The data received as a bytestring
        :rtype: bytes
        :raise SpinnmanTimeoutException:
            If a timeout occurs before any data is received
        :raise SpinnmanIOException: If an error occurs receiving the data
        """
        try:
            self._socket.settimeout(timeout)
            return self._socket.recv(300)
        except socket.timeout as e:
            raise SpinnmanTimeoutException("receive", timeout) from e
        except Exception as e:  # pylint: disable=broad-except
            raise SpinnmanIOException(str(e)) from e

    def receive_with_address(self, timeout=None):
        """ Receive data from the connection along with the address where the\
            data was received from

        :param float timeout: The timeout, or None to wait forever
        :return: A tuple of the data received and a tuple of the
            (address, port) received from
        :rtype: tuple(bytes, tuple(str, int))
        :raise SpinnmanTimeoutException:
            If a timeout occurs before any data is received
        :raise SpinnmanIOException: If an error occurs receiving the data
        """
        try:
            self._socket.settimeout(timeout)
            return self._socket.recvfrom(300)
        except socket.timeout as e:
            raise SpinnmanTimeoutException("receive", timeout) from e
        except Exception as e:  # pylint: disable=broad-except
            raise SpinnmanIOException(str(e)) from e

    def send(self, data):
        """ Send data down this connection

        :param data: The data to be sent
        :type data: bytes or bytearray
        :raise SpinnmanIOException: If there is an error sending the data
        """
        if not self._can_send:
            raise SpinnmanIOException(
                "Remote host and/or port not set - data cannot be sent with"
                " this connection")
        try:
            while not self._socket.send(data):
                pass
        except Exception as e:  # pylint: disable=broad-except
            raise SpinnmanIOException(str(e)) from e

    def send_to(self, data, address):
        """ Send data down this connection

        :param data: The data to be sent as a bytestring
        :type data: bytes or bytearray
        :param tuple(str,int) address:
            A tuple of (address, port) to send the data to
        :raise SpinnmanIOException: If there is an error sending the data
        """
        try:
            while not self._socket.sendto(data, address):
                pass
        except Exception as e:  # pylint: disable=broad-except
            raise SpinnmanIOException(str(e)) from e

    @overrides(Connection.close)
    def close(self):
        try:
            self._socket.shutdown(socket.SHUT_WR)
        except Exception:  # pylint: disable=broad-except
            pass
        self._socket.close()

    def is_ready_to_receive(self, timeout=0):
        return len(select.select([self._socket], [], [], timeout)[0]) == 1

    def __repr__(self):
        return _REPR_TEMPLATE.format(
            self.local_ip_address, self.local_port,
            self.remote_ip_address, self.remote_port)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinnman.messages.scp.impl import SDRAMAlloc
from .abstract_multi_connection_process import AbstractMultiConnectionProcess


class MallocSDRAMProcess(AbstractMultiConnectionProcess):
    """ A process for allocating a block of SDRAM on a SpiNNaker chip.
    """
    __slots__ = [
        "_base_address"]

    def __init__(self, connection_selector):
        """
        :param connection_selector:
        :type connection_selector:
            AbstractMultiConnectionProcessConnectionSelector
        """
        super().__init__(connection_selector)
        self._base_address = None

    def _handle_sdram_alloc_response(self, response):
        self._base_address = response.base_address

    def malloc_sdram(self, x, y, size, app_id, tag):
        """ Allocate space in the SDRAM space.

        :param int x:
        :param int y:
        :param int size:
        :param int app_id:
        :param int tag:
        """
        # pylint: disable=too-many-arguments
        self._send_request(SDRAMAlloc(x, y, app_id, size, tag),
                           self._handle_sdram_alloc_response)
        self._finish()
        self.check_for_error()

    @property
    def base_address(self):
        """ The address of the allocated memory block.

        :rtype: int
        """
        return self._base_address



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.abstract_base import (
    AbstractBase, abstractmethod, abstractproperty)
from .connection import Connection


class SCPSender(Connection, metaclass=AbstractBase):
    """ A sender of SCP messages
    """

    __slots__ = ()

    @abstractmethod
    def get_scp_data(self, scp_request):
        """ Returns the data of an SCP request as it would be sent down this\
            connection
        """

    @abstractmethod
    def send_scp_request(self, scp_request):
        """ Sends an SCP request down this connection

        Messages must have the following properties:

            * source_port is None or 7
            * source_cpu is None or 31
            * source_chip_x is None or 0
            * source_chip_y is None or 0

        tag in the message is optional; if not set, the default set in the\
        constructor will be used.
        sequence in the message is optional; if not set, (sequence number\
        last assigned + 1) % 65536 will be used

        :param AbstractSCPRequest scp_request: message packet to send
        :raise SpinnmanIOException:
            If there is an error sending the message
        """

    @abstractproperty
    def chip_x(self):
        """ The x-coordinate of the chip at which messages sent down this\
            connection will arrive at first

        :rtype: int
        """

    @abstractproperty
    def chip_y(self):
        """ The y-coordinate of the chip at which messages sent down this\
            connection will arrive at first

        :rtype: int
        """



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import struct
from time import localtime, asctime
from spinnman.exceptions import SpinnmanInvalidParameterException

_VERSION_PATTERN = struct.Struct("<BBBBHHI")


class VersionInfo(object):
    """ Decodes SC&MP/SARK version information as returned by the SVER command.
    """
    __slots__ = [
        "_build_date",
        "_hardware",
        "_name",
        "_physical_cpu_id",
        "_version_number",
        "_version_string",
        "_x", "_y", "_p"]

    def __init__(self, version_data, offset=0):
        """
        :param bytes version_data:
            bytes from an SCP packet containing version information
        :param int offset:
            the offset in the bytes from an SCP packet containing
            version information
        :raise SpinnmanInvalidParameterException:
            If the message does not contain valid version information
        """
        (self._p, self._physical_cpu_id, self._y, self._x, _,
            version_no, self._build_date) = _VERSION_PATTERN.unpack_from(
                memoryview(version_data), offset)

        version_data = version_data[offset + 12:-1].decode("utf-8")

        if version_no < 0xFFFF:
            try:
                self._version_number = (version_no // 100, version_no % 100, 0)
                self._name, self._hardware = version_data.split("/")
                self._version_string = version_data
            except ValueError as exception:
                raise SpinnmanInvalidParameterException(
                    "version_data", version_data,
                    "Incorrect format: {}".format(exception)) from exception
        else:
            name_hardware, _, version = version_data.partition("\0")
            self._version_string = version
            matches = re.match(r"(\d+)\.(\d+)\.(\d+)", version)
            if matches is None:
                raise SpinnmanInvalidParameterException(
                    "version", version, "Cannot be parsed")
            self._version_number = tuple(map(int, matches.group(1, 2, 3)))
            self._name, self._hardware = name_hardware.rstrip("\0").split("/")

    @property
    def name(self):
        """ The name of the software.

        :rtype: str
        """
        return self._name

    @property
    def version_number(self):
        """ The version number of the software.

        :rtype: tuple(int, int, int)
        """
        return self._version_number

    @property
    def hardware(self):
        """ The hardware being run on.

        :rtype: str
        """
        return self._hardware

    @property
    def x(self):
        """ The x-coordinate of the chip where the information was obtained.

        :rtype: int
        """
        return self._x

    @property
    def y(self):
        """ The y-coordinate of the chip where the information was obtained.

        :rtype: int
        """
        return self._y

    @property
    def p(self):
        """ The processor ID of the processor where the information was\
            obtained.

        :rtype: int
        """
        return self._p

    @property
    def build_date(self):
        """ The build date of the software, in seconds since 1st January 1970

        :rtype: int
        """
        return self._build_date

    @property
    def version_string(self):
        """ The version information as text.

        :rtype: str
        """
        return self._version_string

    def __str__(self):
        return "[Version: {} {} at {}:{}:{}:{} (built {})]".format(
            self._name, self._version_string, self._hardware, self._x, self._y,
            self._p, asctime(localtime(self._build_date)))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class MulticastMessage(object):
    """ A SpiNNaker Multicast message.
    """
    __slots__ = [
        "_key",
        "_payload"]

    def __init__(self, key, payload=None):
        """ A multicast message has a key (determining the target locations) \
            and an optional payload.

        :param int key: The key of the packet
        :param int payload: The optional payload of the packet
        """
        self._key = key
        self._payload = payload

    @property
    def key(self):
        """ The key of the packet

        :return: The key
        :rtype: int
        """
        return self._key

    @property
    def payload(self):
        """ The payload of the packet if there is one, or None if there is no\
            payload

        :rtype: int or None
        """
        return self._payload



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinnman.messages.scp.impl.fixed_route_read import FixedRouteRead
from .abstract_multi_connection_process import AbstractMultiConnectionProcess


class ReadFixedRouteRoutingEntryProcess(AbstractMultiConnectionProcess):
    """ A process for reading a fixed route routing table entry.
    """

    __slots__ = (
        # the fixed route routing entry from the response
        "_route",
    )

    def __init__(self, connection_selector):
        """
        :param connection_selector: the SC&MP connection selector
        :type connection_selector:
            AbstractMultiConnectionProcessConnectionSelector
        """
        super().__init__(connection_selector)
        self._route = None

    def __handle_read_response(self, response):
        self._route = response.route

    def read_fixed_route(self, x, y, app_id=0):
        """ Read the fixed route entry installed on a particular chip's router.

        :param int x: The x-coordinate of the chip, between 0 and 255;
            this is not checked due to speed restrictions
        :param int y: The y-coordinate of the chip, between 0 and 255;
            this is not checked due to speed restrictions
        :param int app_id:
            The ID of the application with which to associate the
            routes.  If not specified, defaults to 0.
        :rtype: ~spinn_machine.FixedRouteEntry
        """
        self._send_request(FixedRouteRead(x, y, app_id),
                           self.__handle_read_response)
        self._finish()
        self.check_for_error()
        return self._route



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
from .eieio_command_message import EIEIOCommandMessage
from .eieio_command_header import EIEIOCommandHeader
from spinnman.constants import EIEIO_COMMAND_IDS

_PATTERN_BBBxBBI = struct.Struct("<BBBxBBI")


class SpinnakerRequestBuffers(EIEIOCommandMessage):
    """ Message used in the context of the buffering input mechanism which is\
        sent by the SpiNNaker system to the host computer to ask for more data\
        to inject during the simulation
    """
    __slots__ = [
        "_region_id",
        "_sequence_no",
        "_space_available",
        "_p", "_x", "_y"]

    def __init__(self, x, y, p, region_id, sequence_no, space_available):
        # pylint: disable=too-many-arguments
        super().__init__(EIEIOCommandHeader(
            EIEIO_COMMAND_IDS.SPINNAKER_REQUEST_BUFFERS))
        self._x = x
        self._y = y
        self._p = p
        self._region_id = region_id
        self._sequence_no = sequence_no
        self._space_available = space_available

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def p(self):
        return self._p

    @property
    def region_id(self):
        return self._region_id

    @property
    def sequence_no(self):
        return self._sequence_no

    @property
    def space_available(self):
        return self._space_available

    @staticmethod
    def get_min_packet_length():
        return 12

    @staticmethod
    def from_bytestring(command_header, data, offset):  # @UnusedVariable
        y, x, processor, region_id, sequence_no, space = \
            _PATTERN_BBBxBBI.unpack_from(data, offset)
        p = (processor >> 3) & 0x1F
        return SpinnakerRequestBuffers(
            x, y, p, region_id & 0xF, sequence_no, space)

    @property
    def bytestring(self):
        return (super().bytestring + _PATTERN_BBBxBBI.pack(
            self._y, self._x, self._p << 3, self._region_id,
            self._sequence_no, self._space_available))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
from spinnman.constants import SCP_SCAMP_PORT
from spinnman.messages.scp.enums import SCPResult
from .utils import update_sdp_header_for_udp_send
from .sdp_connection import SDPConnection
from spinnman.connections.abstract_classes import SCPSender, SCPReceiver
from spinn_utilities.overrides import overrides

_TWO_SHORTS = struct.Struct("<2H")
_TWO_SKIP = struct.Struct("<2x")
_REPR_TEMPLATE = "SCAMPConnection(chip_x={}, chip_y={}, local_host={}," \
    " local_port={}, remote_host={}, remote_port={})"


class SCAMPConnection(SDPConnection, SCPSender, SCPReceiver):
    """ A UDP connection to SCAMP on the board.
    """
    __slots__ = []

    def __init__(
            self, chip_x=255, chip_y=255, local_host=None, local_port=None,
            remote_host=None, remote_port=None):
        """
        :param int chip_x:
            The x-coordinate of the chip on the board with this remote_host
        :param int chip_y:
            The y-coordinate of the chip on the board with this remote_host
        :param str local_host: The optional IP address or host name of the
            local interface to listen on
        :param int local_port: The optional local port to listen on
        :param str remote_host: The optional remote host name or IP address to
            send messages to.  If not specified, sending will not be possible
            using this connection
        :param int remote_port: The optional remote port number to send
            messages to. If not specified, sending will not be possible using
            this connection
        """
        # pylint: disable=too-many-arguments
        if remote_port is None:
            remote_port = SCP_SCAMP_PORT
        super().__init__(
            chip_x, chip_y, local_host, local_port, remote_host, remote_port)

    @property
    def chip_x(self):
        return self._chip_x

    @property
    def chip_y(self):
        return self._chip_y

    def update_chip_coordinates(self, x, y):
        self._chip_x = x
        self._chip_y = y

    @overrides(SCPSender.get_scp_data,
               additional_arguments=['x', 'y'], extend_defaults=True)
    def get_scp_data(self, scp_request, x=None, y=None):
        """
        :param int x: Optional: x-coordinate of where to send to
        :param int y: Optional: y-coordinate of where to send to
        """
        # pylint: disable=arguments-differ
        if x is None:
            x = self._chip_x
        if y is None:
            y = self._chip_y
        update_sdp_header_for_udp_send(scp_request.sdp_header, x, y)
        return _TWO_SKIP.pack() + scp_request.bytestring

    @overrides(SCPReceiver.receive_scp_response)
    def receive_scp_response(self, timeout=1.0):
        data = self.receive(timeout)
        result, sequence = _TWO_SHORTS.unpack_from(data, 10)
        return SCPResult(result), sequence, data, 2

    def receive_scp_response_with_address(self, timeout=1.0):
        data, (addr, port) = self.receive_with_address(timeout)
        result, sequence = _TWO_SHORTS.unpack_from(data, 10)
        return SCPResult(result), sequence, data, 2, addr, port

    @overrides(SCPSender.send_scp_request)
    def send_scp_request(self, scp_request):
        self.send(self.get_scp_data(scp_request))

    def send_scp_request_to(self, scp_request, x, y, ip_address):
        self.send_to(
            self.get_scp_data(scp_request, x, y),
            (str(ip_address), SCP_SCAMP_PORT))

    def __repr__(self):
        return _REPR_TEMPLATE.format(
            self._chip_x, self._chip_y, self.local_ip_address,
            self.local_port, self.remote_ip_address, self.remote_port)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class CPUState(Enum):
    """ SARK CPU States
    """
    #: Dead core
    DEAD = 0
    #: Powered down
    POWERED_DOWN = 1
    #: Died with Run Time Error
    RUN_TIME_EXCEPTION = 2
    #: Watchdog expired
    WATCHDOG = 3
    #: Initialising (transient)
    INITIALISING = 4
    #: Ready to execute
    READY = 5
    #: Entered c_main
    C_MAIN = 6
    #: Running (API/Event)
    RUNNING = 7
    #: Waiting for sync 0
    SYNC0 = 8
    #: Waiting for sync 1
    SYNC1 = 9
    #: Paused in application
    PAUSED = 10
    #: Exited application
    FINISHED = 11
    #: Spare
    CPU_STATE_12 = 12
    #: Spare
    CPU_STATE_13 = 13
    #: Spare
    CPU_STATE_14 = 14
    #: Idle (SARK stub)
    IDLE = 15

    def __new__(cls, value, doc=""):
        # pylint: disable=protected-access, unused-argument
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .eieio_message import AbstractEIEIOMessage
from .eieio_prefix import EIEIOPrefix
from .eieio_type import EIEIOType
from .create_eieio_command import read_eieio_command_message
from .create_eieio_data import read_eieio_data_message

__all__ = ["EIEIOPrefix", "EIEIOType", "read_eieio_command_message",
           "read_eieio_data_message", "AbstractEIEIOMessage"]



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
from spinn_utilities.overrides import overrides
from spinn_machine import FixedRouteEntry
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, AbstractSCPResponse)
from spinnman.messages.scp.enums import SCPCommand, SCPResult
from spinnman.messages.sdp import SDPHeader, SDPFlag

_ONE_WORD = struct.Struct("<I")


class FixedRouteRead(AbstractSCPRequest):
    """ Gets a fixed route entry
    """

    __slots__ = []

    def __init__(self, x, y, app_id):
        """
        :param int x: The x-coordinate of the chip, between 0 and 255,
            this is not checked due to speed restrictions
        :param int y: The y-coordinate of the chip, between 0 and 255,
            this is not checked due to speed restrictions
        :param int app_id:
            The ID of the application with which to associate the routes.
            If not specified, defaults to 0.
        :raise SpinnmanInvalidParameterException:
            * If x is out of range
            * If y is out of range
        """
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_RTR),
            argument_1=(app_id << 8) | 3, argument_2=1 << 31)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self):
        return _FixedRouteResponse()


class _FixedRouteResponse(AbstractSCPResponse):
    """
    response for the fixed route read
    """
    __slots__ = [
        # the fixed route router entry
        "_route"]

    def __init__(self):
        super().__init__()
        self._route = None

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data, offset):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "Read Fixed Route entry", "CMD_RTR", result.name)

        self._route = _ONE_WORD.unpack_from(data, offset)[0]

    @property
    def route(self):
        processor_ids = list()
        for processor_id in range(0, 26):
            if self._route & (1 << (6 + processor_id)) != 0:
                processor_ids.append(processor_id)
        link_ids = list()
        for link_id in range(0, 6):
            if self._route & (1 << link_id) != 0:
                link_ids.append(link_id)
        return FixedRouteEntry(processor_ids, link_ids)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
from spinnman.exceptions import (
    SpinnmanInvalidPacketException, SpinnmanInvalidParameterTypeException)
from .eieio_command_message import EIEIOCommandMessage
from .eieio_command_header import EIEIOCommandHeader
from spinnman.constants import EIEIO_COMMAND_IDS

_PATTERN_BB = struct.Struct("<BB")
_PATTERN_xxBBI = struct.Struct("<xxBBI")


class HostDataRead(EIEIOCommandMessage):
    """ Packet sent by the host computer to the SpiNNaker system in the\
        context of the buffering output technique to signal that the host has\
        completed reading data from the output buffer, and that such space can\
        be considered free to use again
    """
    __slots__ = [
        "_acks",
        "_header"]

    def __init__(
            self, n_requests, sequence_no, channel, region_id, space_read):
        # pylint: disable=too-many-arguments
        if not isinstance(channel, list):
            channel = [channel]

        if not isinstance(region_id, list):
            region_id = [region_id]

        if not isinstance(space_read, list):
            space_read = [space_read]

        if len(channel) != n_requests or \
                len(region_id) != n_requests or \
                len(space_read) != n_requests:
            raise SpinnmanInvalidPacketException(
                "SpinnakerRequestReadData",
                "The format for a SpinnakerRequestReadData packet is "
                "invalid: {0:d} request(s), {1:d} space(s) read "
                "defined, {2:d} region(s) defined, {3:d} channel(s) "
                "defined".format(
                    n_requests, len(space_read), len(region_id), len(channel)))
        super().__init__(EIEIOCommandHeader(EIEIO_COMMAND_IDS.HOST_DATA_READ))
        self._header = _HostDataReadHeader(n_requests, sequence_no)
        self._acks = _HostDataReadAck(channel, region_id, space_read)

    @property
    def n_requests(self):
        return self._header.n_requests

    @property
    def sequence_no(self):
        return self._header.sequence_no

    def channel(self, ack_id):
        return self._acks.channel(ack_id)

    def region_id(self, ack_id):
        return self._acks.region_id(ack_id)

    def space_read(self, ack_id):
        return self._acks.space_read(ack_id)

    @staticmethod
    def get_min_packet_length():
        return 8

    @staticmethod
    def from_bytestring(command_header, data, offset):  # @UnusedVariable
        n_requests, sequence_no = _PATTERN_BB.unpack_from(data, offset)

        offset += 2
        n_requests &= 0x7
        channel = list()
        region_id = list()
        space_read = list()

        for _ in range(n_requests):
            channel_ack, region_id_ack, space_read_ack = \
                _PATTERN_xxBBI.unpack_from(data, offset)
            channel.append(channel_ack)
            region_id.append(region_id_ack)
            space_read.append(space_read_ack)
            offset += 8
        return HostDataRead(
            n_requests, sequence_no, channel, region_id, space_read)

    @property
    def bytestring(self):
        byte_string = super().bytestring
        n_requests = self.n_requests
        byte_string += _PATTERN_BB.pack(n_requests, self.sequence_no)
        for i in range(n_requests):
            byte_string += _PATTERN_xxBBI.pack(
                self.channel(i), self.region_id(i), self.space_read(i))
        return byte_string


class _HostDataReadHeader(object):
    """ The HostDataRead contains itself on header with the number of requests\
        and a sequence number
    """
    __slots__ = [
        "_n_requests",
        "_sequence_no"]

    def __init__(self, n_requests, sequence_no):
        self._n_requests = n_requests
        self._sequence_no = sequence_no

    @property
    def sequence_no(self):
        return self._sequence_no

    @property
    def n_requests(self):
        return self._n_requests


class _HostDataReadAck(object):
    """ Contains a set of acks which refer to each of the channels read
    """
    __slots__ = [
        "_channel",
        "_region_id",
        "_space_read"]

    def __init__(self, channel, region_id, space_read):
        if not isinstance(channel, list):
            self._channel = [channel]
        else:
            self._channel = channel

        if not isinstance(region_id, list):
            self._region_id = [region_id]
        else:
            self._region_id = region_id

        if not isinstance(space_read, list):
            self._space_read = [space_read]
        else:
            self._space_read = space_read

    def channel(self, ack_id):
        if len(self._channel) > ack_id:
            return self._channel[ack_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer", "channel ack_id needs to be"
            "comprised between 0 and {0:d}; current value: {1:d}".format(
                len(self._channel) - 1, ack_id))

    def region_id(self, ack_id):
        if len(self._region_id) > ack_id:
            return self._region_id[ack_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer", "region ID ack_id needs to be"
            "comprised between 0 and {0:d}; current value: {1:d}".format(
                len(self._region_id) - 1, ack_id))

    def space_read(self, ack_id):
        if len(self._space_read) > ack_id:
            return self._space_read[ack_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer", "start address ack_id needs to be"
            "comprised between 0 and {0:d}; current value: {1:d}".format(
                len(self._space_read) - 1, ack_id))



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import traceback
from collections import OrderedDict


class SpinnmanException(Exception):
    """ Superclass of exceptions that occur when dealing with communication\
        with SpiNNaker
    """


class SpinnmanInvalidPacketException(SpinnmanException):
    """ An exception that indicates that a packet was not in the expected\
        format
    """

    def __init__(self, packet_type, problem):
        """
        :param str packet_type: The type of packet expected
        :param str problem: The problem with the packet
        """
        super().__init__("Invalid packet of type {} received: {}".format(
            packet_type, problem))
        self._packet_type = packet_type
        self._problem = problem

    @property
    def packet_type(self):
        """ The packet type
        """
        return self._packet_type

    @property
    def problem(self):
        """ The problem with the packet
        """
        return self._problem


class SpinnmanInvalidParameterException(SpinnmanException):
    """ An exception that indicates that the value of one of the parameters\
        passed was invalid
    """

    def __init__(self, parameter, value, problem):
        """
        :param str parameter: The name of the parameter that is invalid
        :param str value: The value of the parameter that is invalid
        :param str problem: The problem with the parameter
        """
        super().__init__(
            "Setting parameter {} to value {} is invalid: {}".format(
                parameter, value, problem))
        self._parameter = parameter
        self._value = value
        self._problem = problem

    @property
    def parameter(self):
        """ The parameter with an invalid value
        """
        return self._parameter

    @property
    def value(self):
        """ The value that is invalid
        """
        return self._value

    @property
    def problem(self):
        """ The problem with the parameter value
        """
        return self._problem


class SpinnmanInvalidParameterTypeException(SpinnmanException):
    """ An exception that indicates that the type of one of the parameters\
        passed was invalid
    """

    def __init__(self, parameter, param_type, problem):
        """
        :param str parameter: The name of the parameter that is invalid
        :param str param_type: The type of the parameter that is invalid
        :param str problem: The problem with the parameter
        """
        super().__init__("Parameter {} of type {} is invalid: {}".format(
            parameter, param_type, problem))
        self._parameter = parameter
        self._type = param_type
        self._problem = problem

    @property
    def parameter(self):
        """ The parameter with an invalid value
        """
        return self._parameter

    @property
    def type(self):
        """ The value that is invalid
        """
        return self._type

    @property
    def problem(self):
        """ The problem with the parameter value
        """
        return self._problem


class SpinnmanIOException(SpinnmanException):
    """ An exception that something went wrong with the underlying IO
    """

    def __init__(self, problem):
        """
        :param str problem: The problem with the IO
        """
        super().__init__("IO Error: {}".format(problem))
        self._problem = problem

    @property
    def problem(self):
        """ The problem with IO
        """
        return self._problem


class SpinnmanTimeoutException(SpinnmanException):
    """ An exception that indicates that a timeout occurred before an operation
        could finish
    """

    def __init__(self, operation, timeout, msg=None):
        """
        :param str operation: The operation being performed
        :param float timeout: The timeout value in seconds
        """
        if msg is None:
            msg = "Operation {} timed out after {} seconds".format(
                operation, timeout)
        super().__init__(msg)

        self._operation = operation
        self._timeout = timeout

    @property
    def operation(self):
        """ The operation that was performed
        """
        return self._operation

    @property
    def timeout(self):
        """ The timeout value in seconds
        """
        return self._timeout


class SpinnmanUnexpectedResponseCodeException(SpinnmanException):
    """ Indicate that a response code returned from the board was unexpected\
        for the current operation
    """

    def __init__(self, operation, command, response):
        """
        :param str operation: The operation being performed
        :param str command: The command being executed
        :param str response: The response received in error
        """
        super().__init__(
            "Unexpected response {} while performing operation {} using"
            " command {}".format(response, operation, command))
        self._operation = operation
        self._command = command
        self._response = response

    @property
    def operation(self):
        """ The operation being performed
        """
        return self._operation

    @property
    def command(self):
        """ The command being executed
        """
        return self._command

    @property
    def response(self):
        """ The unexpected response
        """
        return self._response


class _Group(object):
    def __init__(self, trace_back, connection):
        self.trace_back = trace_back
        self.chip_core = "board {} with ethernet chip {}:{} [".format(
            connection.remote_ip_address, connection.chip_x, connection.chip_y)
        self._separator = ""

    def finalise(self):
        self.chip_core += "]"

    def add_coord(self, sdp_header):
        self.chip_core += "{}[{}:{}:{}]".format(
            self._separator,
            sdp_header.destination_chip_x,
            sdp_header.destination_chip_y,
            sdp_header.destination_cpu)
        self._separator = ","

    @staticmethod
    def group_exceptions(error_requests, exceptions, tracebacks, connections):
        """ Groups exceptions into a form usable by an exception.

        :param list(SCPRequest) error_requests: the error requests
        :param list(Exception) exceptions: the exceptions
        :param list tracebacks: the tracebacks
        :param list connections:
            the connections the errors were associated with
        :return: a sorted exception pile
        :rtype: dict(Exception,_Group)
        """
        data = OrderedDict()
        for error_request, exception, trace_back, connection in zip(
                error_requests, exceptions, tracebacks, connections):
            for stored_exception in data.keys():
                if isinstance(exception, type(stored_exception)):
                    found_exception = stored_exception
                    break
            else:
                data[exception] = _Group(trace_back, connection)
                found_exception = exception
            data[found_exception].add_coord(error_request.sdp_header)
        for exception in data:
            data[exception].finalise()
        return data.items()


class SpinnmanGroupedProcessException(SpinnmanException):
    """ Encapsulates exceptions from processes which communicate with a\
        collection of cores/chips
    """
    def __init__(self, error_requests, exceptions, tracebacks, connections):
        problem = "Exceptions found were:\n"
        for exception, description in _Group.group_exceptions(
                error_requests, exceptions, tracebacks, connections):
            problem += \
                "   Received exception class: {}\n" \
                "       With message {}\n" \
                "       When sending to {}\n" \
                "       Stack trace: {}\n".format(
                    exception.__class__.__name__, str(exception),
                    description.chip_core,
                    traceback.format_tb(description.trace_back))
        super().__init__(problem)


class SpinnmanGenericProcessException(SpinnmanException):
    """ Encapsulates exceptions from processes which communicate with some\
        core/chip
    """
    def __init__(self, exception, tb, x, y, p, tb2=None):
        """
        :param Exception exception:
        :param int x:
        :param int y:
        :param int p:
        """
        # pylint: disable=too-many-arguments
        super().__init__(
            "\n     Received exception class: {} \n"
            "     With message: {} \n"
            "     When sending to {}:{}:{}\n"
            "     Stack trace: {}\n".format(
                exception.__class__.__name__, str(exception), x, y, p,
                traceback.format_tb(tb)))

        self._stored_exception = exception
        if tb2 is not None:
            self.__traceback__ = tb2

    @property
    def exception(self):
        """
        :rtype: Exception
        """
        return self._stored_exception


class SpinnmanUnsupportedOperationException(SpinnmanException):
    """ An exception that indicates that the given operation is not supported
    """

    def __init__(self, operation):
        """
        :param str operation: The operation being requested
        """
        super().__init__("Operation {} is not supported".format(operation))
        self._operation = operation

    @property
    def operation(self):
        """ The unsupported operation requested

        :rtype: str
        """
        return self._operation


class SpinnmanEIEIOPacketParsingException(SpinnmanException):
    """ Unable to complete the parsing of the EIEIO packet received.\
    The routine used is invalid or the content of the packet is invalid
    """

    def __init__(self, parsing_format, packet):
        """
        :param str parsing_format:
        :param bytes packet:
        """
        super().__init__(
            "The packet received is being parsed as an EIEIO {0:s} packet, "
            "but the content of the packet is invalid".format(parsing_format))
        self._packet = packet

    @property
    def packet(self):
        """
        :rtype: bytes
        """
        return self._packet


class SpiNNManCoresNotInStateException(SpinnmanTimeoutException):
    """ Cores failed to reach a given state within a timeout.
    """

    def __init__(self, timeout, expected_states, failed_core_states):
        """
        :param float timeout:
        :param set(CPUState) expected_states:
        :param CPUInfos failed_core_states:
        """

        msg = "waiting for cores {} to reach one of {}".format(
            failed_core_states, expected_states)
        super().__init__(msg, timeout, msg)
        self._failed_core_states = failed_core_states

    def failed_core_states(self):
        """
        :rtype: CPUInfos
        """
        return self._failed_core_states



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class RunTimeError(Enum):
    """ SARK Run time errors
    """
    #: No error
    NONE = 0
    #: Branch through zero
    RESET = 1
    #: Undefined instruction
    UNDEF = 2
    #: Undefined SVC or no handler
    SVC = 3
    #: Prefetch abort
    PABT = 4
    #: Data abort
    DABT = 5
    #: Unhandled IRQ
    IRQ = 6
    #: Unhandled FIQ
    FIQ = 7
    #: Unconfigured VIC vector
    VIC = 8
    #: Generic user abort
    ABORT = 9
    #: "malloc" failure
    MALLOC = 10
    #: Divide by zero
    DIVBY0 = 11
    #: Event startup failure
    EVENT = 12
    #: Fatal SW error
    SWERR = 13
    #: Failed to allocate IO buffer
    IOBUF = 14
    #: Bad event enable
    ENABLE = 15
    #: Generic null pointer error
    NULL = 16
    #: Pkt startup failure
    PKT = 17
    #: Timer startup failure
    TIMER = 18
    #: API startup failure
    API = 19
    #: SW version conflict
    SARK_VERSRION_INCORRECT = 20

    def __new__(cls, value, doc=""):
        # pylint: disable=protected-access, unused-argument
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
from spinn_utilities.overrides import overrides
from spinnman.messages.sdp import SDPMessage, SDPFlag
from .udp_connection import UDPConnection
from .utils import update_sdp_header_for_udp_send
from spinnman.connections.abstract_classes import (
    SDPReceiver, SDPSender, Listenable)

_TWO_SKIP = struct.Struct("<2x")
_REPR_TEMPLATE = "SDPConnection(chip_x={}, chip_y={}, local_host={},"\
    " local_port={}, remote_host={}, remote_port={})"


class SDPConnection(UDPConnection, SDPReceiver, SDPSender, Listenable):
    __slots__ = [
        "_chip_x",
        "_chip_y"]

    def __init__(self, chip_x=None, chip_y=None, local_host=None,
                 local_port=None, remote_host=None, remote_port=None):
        """
        :param int chip_x: The optional x-coordinate of the chip at the remote
            end of the connection. If not specified, it will not be possible
            to send SDP messages that require a response with this connection.
        :param int chip_y: The optional y-coordinate of the chip at the remote
            end of the connection. If not specified, it will not be possible
            to send SDP messages that require a response with this connection.
        :param str local_host: The optional IP address or host name of the
            local interface to listen on
        :param int local_port: The optional local port to listen on
        :param str remote_host: The optional remote host name or IP address to
            send messages to. If not specified, sending will not be possible
            using this connection
        :param int remote_port: The optional remote port number to send
            messages to. If not specified, sending will not be possible using
            this connection
        """
        # pylint: disable=too-many-arguments
        super().__init__(local_host, local_port, remote_host, remote_port)
        self._chip_x = chip_x
        self._chip_y = chip_y

    @overrides(SDPReceiver.receive_sdp_message)
    def receive_sdp_message(self, timeout=None):
        data = self.receive(timeout)
        return SDPMessage.from_bytestring(data, 2)

    @overrides(SDPSender.send_sdp_message)
    def send_sdp_message(self, sdp_message):
        # If a reply is expected, the connection should
        if sdp_message.sdp_header.flags == SDPFlag.REPLY_EXPECTED:
            update_sdp_header_for_udp_send(
                sdp_message.sdp_header, self._chip_x, self._chip_y)
        else:
            update_sdp_header_for_udp_send(sdp_message.sdp_header, 0, 0)
        self.send(_TWO_SKIP.pack() + sdp_message.bytestring)

    @overrides(Listenable.get_receive_method)
    def get_receive_method(self):
        return self.receive_sdp_message

    def __repr__(self):
        return _REPR_TEMPLATE.format(
            self._chip_x, self._chip_y, self.local_ip_address,
            self.local_port, self.remote_ip_address, self.remote_port)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from .count_state_response import CountStateResponse

_ALL_CORE_MASK = 0xFFFF
_COUNT_OPERATION = 1
_COUNT_MODE = 2
_COUNT_SIGNAL_TYPE = 1
_APP_MASK = 0xFF


def _get_data(app_id, state):
    data = (_APP_MASK << 8) | app_id
    data += (_COUNT_OPERATION << 22) | (_COUNT_MODE << 20)
    data += state.value << 16
    return data


class CountState(AbstractSCPRequest):
    """ An SCP Request to get a count of the cores in a particular state
    """
    __slots__ = []

    def __init__(self, app_id, state):
        """

        :param int app_id: The ID of the application, between 0 and 255
        :param CPUState state: The state to count
        """
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0,
                destination_chip_x=self.DEFAULT_DEST_X_COORD,
                destination_chip_y=self.DEFAULT_DEST_Y_COORD),
            SCPRequestHeader(command=SCPCommand.CMD_SIG),
            argument_1=_COUNT_SIGNAL_TYPE,
            argument_2=_get_data(app_id, state),
            argument_3=_ALL_CORE_MASK)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self):
        return CountStateResponse()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
from .udp_connection import UDPConnection
from spinnman.connections.abstract_classes import (
    EIEIOReceiver, EIEIOSender, Listenable)
from spinnman.messages.eieio import (
    read_eieio_command_message, read_eieio_data_message)
from spinn_utilities.overrides import overrides

_ONE_SHORT = struct.Struct("<H")
_REPR_TEMPLATE = "EIEIOConnection(local_host={}, local_port={},"\
    "remote_host={}, remote_port={})"


class EIEIOConnection(
        UDPConnection, EIEIOReceiver, EIEIOSender, Listenable):
    """ A UDP connection for sending and receiving raw EIEIO messages.
    """
    __slots__ = []

    @overrides(EIEIOReceiver.receive_eieio_message)
    def receive_eieio_message(self, timeout=None):
        data = self.receive(timeout)
        header = _ONE_SHORT.unpack_from(data)[0]
        if header & 0xC000 == 0x4000:
            return read_eieio_command_message(data, 0)
        return read_eieio_data_message(data, 0)

    @overrides(EIEIOSender.send_eieio_message)
    def send_eieio_message(self, eieio_message):
        self.send(eieio_message.bytestring)

    def send_eieio_message_to(self, eieio_message, ip_address, port):
        self.send_to(eieio_message.bytestring, (ip_address, port))

    @overrides(Listenable.get_receive_method)
    def get_receive_method(self):
        return self.receive_eieio_message

    def __repr__(self):
        return _REPR_TEMPLATE.format(
            self.local_ip_address, self.local_port,
            self.remote_ip_address, self.remote_port)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, AbstractSCPResponse)
from spinnman.messages.scp.enums import SCPCommand, SCPResult
from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException

import struct


_IPTAG_GET = 2

_IPTAG_FORMAT = struct.Struct("<4s6s3HIH3B")


class IPTagGet(AbstractSCPRequest):
    """ An SCP Request to get an IP tag
    """
    __slots__ = []

    def __init__(self, x, y, tag):
        """
        :param int x: The x-coordinate of a chip, between 0 and 255
        :param int y: The y-coordinate of a chip, between 0 and 255
        :param int tag: The tag to get details of, between 0 and 7
        :param int tag: The tag, between 0 and 7
        """
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_IPTAG),
            argument_1=(_IPTAG_GET << 16) | tag,
            argument_2=1)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self):
        return _SCPIPTagGetResponse()


class _SCPIPTagGetResponse(AbstractSCPResponse):
    """ An SCP response to a request for an IP tags
    """
    __slots__ = [
        "_count",
        "_flags",
        "_ip_address",
        "_mac_address",
        "_port",
        "_rx_port",
        "_spin_chip_x",
        "_spin_chip_y",
        "_spin_cpu",
        "_spin_port",
        "_timeout"]

    def __init__(self):
        """
        """
        super().__init__()
        self._ip_address = None
        self._mac_address = None
        self._port = None
        self._timeout = None
        self._flags = None
        self._count = None
        self._rx_port = None
        self._spin_chip_y = None
        self._spin_chip_x = None
        self._spin_cpu = None
        self._spin_port = None

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data, offset):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "Get IP Tag Info", "CMD_IPTAG", result.name)
        (ip_address, mac_address, self._port, self._timeout,
         self._flags, self._count, self._rx_port, self._spin_chip_y,
         self._spin_chip_x, processor_and_port) = _IPTAG_FORMAT.unpack_from(
            data, offset)
        self._ip_address = bytearray(ip_address)
        self._mac_address = bytearray(mac_address)
        self._spin_port = (processor_and_port >> 5) & 0x7
        self._spin_cpu = processor_and_port & 0x1F

    @property
    def ip_address(self):
        """ The IP address of the tag, as a bytearray of 4 bytes

        :rtype: bytearray
        """
        return self._ip_address

    @property
    def mac_address(self):
        """ The MAC address of the tag, as a bytearray of 6 bytes

        :rtype: bytearray
        """
        return self._mac_address

    @property
    def port(self):
        """ The port of the tag

        :rtype: int
        """
        return self._port

    @property
    def timeout(self):
        """ The timeout of the tag

        :rtype: int
        """
        return self._timeout

    @property
    def flags(self):
        """ The flags of the tag

        :rtype: int
        """
        return self._flags

    @property
    def in_use(self):
        """ True if the tag is marked as being in use

        :rtype: bool
        """
        return (self._flags & 0x8000) > 0

    @property
    def is_temporary(self):
        """ True if the tag is temporary

        :rtype: bool
        """
        return (self._flags & 0x4000) > 0

    @property
    def is_arp(self):
        """ True if the tag is in the ARP state (where the MAC address is\
            being looked up - transient state so unlikely)

        :rtype: bool
        """
        return (self._flags & 0x2000) > 0

    @property
    def is_reverse(self):
        """ True if the tag is a reverse tag

        :rtype: bool
        """
        return (self._flags & 0x0200) > 0

    @property
    def strip_sdp(self):
        """ True if the tag is to strip the SDP header

        :rtype: bool
        """
        return (self._flags & 0x0100) > 0

    @property
    def count(self):
        """ The count of the number of packets that have been sent with the tag

        :rtype: int
        """
        return self._count

    @property
    def rx_port(self):
        """ The receive port of the tag

        :rtype: int
        """
        return self._rx_port

    @property
    def spin_chip_x(self):
        """ The x-coordinate of the chip on which the tag is defined

        :rtype: int
        """
        return self._spin_chip_x

    @property
    def spin_chip_y(self):
        """ The y-coordinate of the chip on which the tag is defined

        :rtype: int
        """
        return self._spin_chip_y

    @property
    def spin_port(self):
        """ The spin-port of the IP tag

        :rtype: int
        """
        return self._spin_port

    @property
    def spin_cpu(self):
        """ The CPU ID of the IP tag

        :rtype: int
        """
        return self._spin_cpu



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
from enum import Enum
from spinnman.exceptions import SpinnmanInvalidParameterException

_ONE_SHORT = struct.Struct("<H")


class EIEIOCommandHeader(object):
    """ EIEIO header for command packets
    """
    __slots__ = [
        "_command"]

    def __init__(self, command):
        if isinstance(command, Enum):
            command = command.value
        if command < 0 or command >= 16384:
            raise SpinnmanInvalidParameterException(
                "command", command,
                "parameter command is outside the allowed range (0 to 16383)")
        self._command = command

    @property
    def command(self):
        return self._command

    @staticmethod
    def from_bytestring(data, offset):
        """ Read an EIEIO command header from a bytestring

        :param data: The bytestring to read the data from
        :type data: bytes or bytearray
        :param int offset: The offset where the valid data starts
        :return: an EIEIO command header
        :rtype: EIEIOCommandHeader
        :raise SpinnmanIOException:
            If there is an error reading from the reader
        :raise SpinnmanInvalidParameterException:
            If there is an error setting any of the values
        """
        command_header = _ONE_SHORT.unpack_from(data, offset)[0]
        command = command_header & 0x3FFF

        return EIEIOCommandHeader(command)

    @property
    def bytestring(self):
        """ Get a bytestring of the header

        :rtype: bytes
        """
        return _ONE_SHORT.pack(0 << 15 | 1 << 14 | self._command)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
from spinnman import constants

_PATTERN = struct.Struct(
    "<"   # Little-endian
    "8H"  # uint16_t adc[8]
    "4h"  # int16_t t_int[4]
    "4h"  # int16_t t_ext[4]
    "4h"  # int16_t fan[4]
    "I"   # uint32_t warning
    "I")  # uint32_t shutdown


class ADCInfo(object):
    """ Container for the ADC data thats been retrieved from an FPGA.
    """
    __slots__ = [
        "_fan_0",
        "_fan_1",
        "_temp_btm",
        "_temp_ext_0",
        "_temp_ext_1",
        "_temp_top",
        "_voltage_1_2a",
        "_voltage_1_2b",
        "_voltage_1_2c",
        "_voltage_1_8",
        "_voltage_3_3",
        "_voltage_supply"]

    def __init__(self, adc_data, offset):
        """
        :param bytes adc_data:
            bytes from an SCP packet containing ADC information
        :raise SpinnmanInvalidParameterException:
            If the message does not contain valid ADC information
        """
        data = _PATTERN.unpack_from(adc_data, offset)

        self._voltage_1_2c = data[1] * constants.BMP_V_SCALE_2_5
        self._voltage_1_2b = data[2] * constants.BMP_V_SCALE_2_5
        self._voltage_1_2a = data[3] * constants.BMP_V_SCALE_2_5
        self._voltage_1_8 = data[4] * constants.BMP_V_SCALE_2_5
        self._voltage_3_3 = data[6] * constants.BMP_V_SCALE_3_3
        self._voltage_supply = data[7] * constants.BMP_V_SCALE_12
        self._temp_top = float(data[8]) * constants.BMP_TEMP_SCALE
        self._temp_btm = float(data[9]) * constants.BMP_TEMP_SCALE
        if data[12] != constants.BMP_MISSING_TEMP:
            self._temp_ext_0 = (float(data[12]) * constants.BMP_TEMP_SCALE)
        else:
            self._temp_ext_0 = None
        if data[13] != constants.BMP_MISSING_TEMP:
            self._temp_ext_1 = (float(data[13]) * constants.BMP_TEMP_SCALE)
        else:
            self._temp_ext_1 = None
        if data[16] != constants.BMP_MISSING_FAN:
            self._fan_0 = float(data[16])
        else:
            self._fan_0 = None
        if data[17] != constants.BMP_MISSING_FAN:
            self._fan_1 = float(data[17])
        else:
            self._fan_1 = None

    @property
    def voltage_1_2c(self):
        """ Actual voltage of the 1.2V c supply rail

        :rtype: float
        """
        return self._voltage_1_2c

    @property
    def voltage_1_2b(self):
        """ Actual voltage of the 1.2V b supply rail

        :rtype: float
        """
        return self._voltage_1_2b

    @property
    def voltage_1_2a(self):
        """ Actual voltage of the 1.2V a supply rail

        :rtype: float
        """
        return self._voltage_1_2a

    @property
    def voltage_1_8(self):
        """ Actual voltage of the 1.8V supply rail

        :rtype: float
        """
        return self._voltage_1_8

    @property
    def voltage_3_3(self):
        """ Actual voltage of the 3.3V supply rail

        :rtype: float
        """
        return self._voltage_3_3

    @property
    def voltage_supply(self):
        """ Actual voltage of the main power supply (nominally 12V).

        :rtype: float
        """
        return self._voltage_supply

    @property
    def temp_top(self):
        """ temperature top

        :rtype: float
        """
        return self._temp_top

    @property
    def temp_btm(self):
        """ temperature bottom

        :rtype: float
        """
        return self._temp_btm

    @property
    def temp_ext_0(self):
        """ temperature external 0

        :rtype: float or None
        """
        return self._temp_ext_0

    @property
    def temp_ext_1(self):
        """ temperature external 1

        :rtype: float or None
        """
        return self._temp_ext_1

    @property
    def fan_0(self):
        """ fan 0

        :rtype: float or None
        """
        return self._fan_0

    @property
    def fan_1(self):
        """ fan 1

        :rtype: float or None
        """
        return self._fan_1



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Used to communicate with a SpiNNaker Board. The main part of this package is
the :py:class:`~spinnman.transceiver.Transceiver` class. This can be used to
send and receive packets in various SpiNNaker formats, depending on what
connections are available.

Functional Requirements
=======================

#. *Connect* to and communicate with a machine using a number of different
   connections.

#. *Boot* a machine with the expected version of the software.

   * If the machine is already booted but the version is not the version
     expected, an exception will be thrown.

#. *Check the version* of the software which the machine is booted with.

#. *Query the state* of the machine to determine:

   * What the current state of the machine is in terms of the chips and cores
     available, the SDRAM available on the chips and which links are available
     between which chips.

   * What external links to the host exist (and separately add the discovered
     links to the set of links used to communicate with the machine).

   * What is running on the machine and where, and what the current status of
     those processes are.

   * How many cores are in a given state.

   * What is in the IOBUF buffers.

   * What the current routing entries for a given router are.

   * What the routing status counter values are.

#. *Load application binaries* on to the machine, either to individual cores or
   via a "flood-fill" mechanism to multiple cores simultaneously (which may be
   a subset of the cores on a subset of the chips).

#. *Write data to SDRAM*, either on an individual chip, or via a "flood-fill"
   mechanism to multiple chips simultaneously.

#. Send a *signal* to an application.

#. *Read data from SDRAM* on an individual chip.

#. *Send and receive SpiNNaker packets* where the connections allow this.

   * If no connection supports this packet type, an exception is thrown.

   * The user should be able to select which connection is used. Selection of a
     connection which does not support the traffic type will also result in an
     exception.

#. *Send and receive SCP and SDP* packets where the connections allow this.

   * If no connection supports the packet type, an exception is thrown.

   * The user should be able to select which connection is used. Selection of a
     connection which does not support the traffic type will also result in an
     exception.

#. It should be possible to *call* any of the functions *simultaneously,*
   including the same function more than once.

   * Where possible, multiple connections should be used to overlap calls.

   * The functions should not return until they have confirmed that any
     messages sent have been received, and any responses have been received.

   * Functions should not respond with the result of a different function.

   * Functions can further sub-divide the call into a number of separate calls
     that can be divided across the available connections, so long as the other
     requirements are met.

#. *More than one machine* can be connected to the same host.

   * Once the subset of connections has been worked out for each machine, the
     operation of these machines should be independent.

Use Cases
=========

* Connecting is done by using
  :py:func:`~spinnman.transceiver.create_transceiver_from_hostname`.

* :py:meth:`~spinnman.transceiver.Transceiver.boot_board` and
  :py:meth:`~spinnman.transceiver.Transceiver.get_scamp_version` are used to
  ensure that the board is booted correctly before starting a simulation.

* :py:meth:`~spinnman.transceiver.Transceiver.get_machine_details` is used to
  get a representation of the current state of the machine, which is used to
  decide where executables are to be run on the board for a particular
  simulation, where any external peripherals are connected, and how messages
  between the executables and/or the external peripherals are to be routed.

* :py:meth:`~spinnman.transceiver.Transceiver.write_memory` and
  :py:meth:`~spinnman.transceiver.Transceiver.execute` are used to write
  parameters and execute executables on the board

* :py:meth:`~spinnman.transceiver.Transceiver.send_signal` is used to send a
  signal which starts, stops or pauses a simulation.

* :py:meth:`~spinnman.transceiver.Transceiver.get_core_state_count` is used to
  determine if a simulation is complete or has gone into an error state.

* :py:meth:`~spinnman.transceiver.Transceiver.get_iobuf`,
  :py:meth:`~spinnman.transceiver.Transceiver.get_cpu_information` and
  :py:meth:`~spinnman.transceiver.Transceiver.get_router_diagnostics` are used
  to diagnose a problem with a simulation.

* :py:meth:`~spinnman.transceiver.Transceiver.read_memory` is used to read some
  statistics recorded in SDRAM after a simulation.
"""
from spinnman._version import __version__  # NOQA
from spinnman._version import __version_name__  # NOQA
from spinnman._version import __version_month__  # NOQA
from spinnman._version import __version_year__  # NOQA



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinnman.model.enums import (
    DiagnosticFilterDestination, DiagnosticFilterSource,
    DiagnosticFilterPayloadStatus, DiagnosticFilterDefaultRoutingStatus,
    DiagnosticFilterEmergencyRoutingStatus, DiagnosticFilterPacketType)

# Bit offsets of the various fields in the filter word
_PACKET_TYPE_OFFSET = 0
_EMERGENCY_ROUTE_OFFSET = 4
_EMERGENCY_ROUTE_MODE_OFFSET = 8
_DEFAULT_ROUTE_OFFSET = 10
_PAYLOAD_OFFSET = 12
_SOURCE_OFFSET = 14
_DESTINATION_OFFSET = 16
_ENABLE_INTERRUPT_OFFSET = 30


# Uses an enum to set flags in the filter word from a given offset
def _set_flags_in_word(word, enum_list, enum_type, offset):
    if enum_list is None:
        enum_values = list()
    else:
        enum_values = list(enum_list)
    if not enum_values:
        enum_values = [value for value in enum_type]
    for enum_value in enum_values:
        word |= 1 << (enum_value.value + offset)
    return word


# Uses an enum to read flags in the filter word from a given offset
def _read_flags_from_word(word, enum_list, offset):
    flags = list()
    for enum_value in enum_list:
        if word & 1 << (enum_value.value + offset) != 0:
            flags.append(enum_value)
    return flags


class DiagnosticFilter(object):
    """ A router diagnostic counter filter, which counts packets passing\
        through the router with certain properties.  The counter will be\
        incremented so long as the packet matches one of the values in each\
        field i.e. one of each of the destinations, sources, payload_statuses,\
        default_routing_statuses, emergency_routing_statuses and packet_types.
    """
    __slots__ = [
        "_default_routing_statuses",
        "_destinations",
        "_emergency_routing_statuses",
        "_enable_interrupt_on_counter_event",
        "_match_emergency_routing_status_to_incoming_packet",
        "_packet_types",
        "_payload_statuses",
        "_sources"]

    def __init__(self, enable_interrupt_on_counter_event,
                 match_emergency_routing_status_to_incoming_packet,
                 destinations, sources, payload_statuses,
                 default_routing_statuses, emergency_routing_statuses,
                 packet_types):
        """
        :param bool enable_interrupt_on_counter_event: Indicates whether
            an interrupt should be raised when this rule matches
        :param bool match_emergency_routing_status_to_incoming_packet:
            Indicates whether the emergency routing statuses should be matched
            against packets arriving at this router (if True), or if they
            should be matched against packets leaving this router (if False)
        :param list(DiagnosticFilterDestination) destinations:
            Increment the counter if one or more of the given destinations
            match
        :param list(DiagnosticFilterSource) sources:
            Increment the counter if one or more of the given sources match
            (or None or empty list to match all)
        :param list(DiagnosticFilterPayloadStatus) payload_statuses:
            Increment the counter if one or more of the given payload statuses
            match (or None or empty list to match all)
        :param default_routing_statuses:
            Increment the counter if one or more of the given default routing
            statuses match (or None or empty list to match all)
        :type default_routing_statuses:
            list(DiagnosticFilterDefaultRoutingStatus)
        :param emergency_routing_statuses:
            Increment the counter if one or more of the given emergency routing
            statuses match (or None or empty list to match all)
        :type emergency_routing_statuses:
            list(DiagnosticFilterEmergencyRoutingStatus)
        :param list(DiagnosticFilterPacketType) packet_types:
            Increment the counter if one or more of the given packet types
            match (or None or empty list to match all)
        """
        # pylint: disable=too-many-arguments
        self._enable_interrupt_on_counter_event = \
            enable_interrupt_on_counter_event
        self._match_emergency_routing_status_to_incoming_packet = \
            match_emergency_routing_status_to_incoming_packet
        self._destinations = destinations
        self._sources = sources
        self._payload_statuses = payload_statuses
        self._default_routing_statuses = default_routing_statuses
        self._emergency_routing_statuses = emergency_routing_statuses
        self._packet_types = packet_types

    @property
    def enable_interrupt_on_counter_event(self):
        return self._enable_interrupt_on_counter_event

    @property
    def match_emergency_routing_status_to_incoming_packet(self):
        return self._match_emergency_routing_status_to_incoming_packet

    @property
    def destinations(self):
        return self._destinations

    @property
    def sources(self):
        return self._sources

    @property
    def payload_statuses(self):
        return self._payload_statuses

    @property
    def default_routing_statuses(self):
        return self._default_routing_statuses

    @property
    def emergency_routing_statuses(self):
        return self._emergency_routing_statuses

    @property
    def packet_types(self):
        return self._packet_types

    @property
    def filter_word(self):
        """ A word of data that can be written to the router to set up\
            the filter
        """
        data = 0
        if self._enable_interrupt_on_counter_event:
            data |= 1 << _ENABLE_INTERRUPT_OFFSET
        if not self._match_emergency_routing_status_to_incoming_packet:
            data |= 1 << _EMERGENCY_ROUTE_MODE_OFFSET
        data = _set_flags_in_word(data, self._destinations,
                                  DiagnosticFilterDestination,
                                  _DESTINATION_OFFSET)
        data = _set_flags_in_word(data, self._sources,
                                  DiagnosticFilterSource,
                                  _SOURCE_OFFSET)
        data = _set_flags_in_word(data, self._payload_statuses,
                                  DiagnosticFilterPayloadStatus,
                                  _PAYLOAD_OFFSET)
        data = _set_flags_in_word(data, self._default_routing_statuses,
                                  DiagnosticFilterDefaultRoutingStatus,
                                  _DEFAULT_ROUTE_OFFSET)
        data = _set_flags_in_word(data, self._emergency_routing_statuses,
                                  DiagnosticFilterEmergencyRoutingStatus,
                                  _EMERGENCY_ROUTE_OFFSET)
        data = _set_flags_in_word(data, self._packet_types,
                                  DiagnosticFilterPacketType,
                                  _PACKET_TYPE_OFFSET)

        return data

    @staticmethod
    def read_from_int(int_value):
        enable_interrupt_on_counter_event = (
            (int_value >> _ENABLE_INTERRUPT_OFFSET) & 0x1) == 1
        match_emergency_routing_status_to_incoming_packet = (
            (int_value >> _EMERGENCY_ROUTE_MODE_OFFSET) & 0x1) == 0
        destinations = _read_flags_from_word(
            int_value, DiagnosticFilterDestination, _DESTINATION_OFFSET)
        sources = _read_flags_from_word(
            int_value, DiagnosticFilterSource, _SOURCE_OFFSET)
        payload_statuses = _read_flags_from_word(
            int_value, DiagnosticFilterPayloadStatus, _PAYLOAD_OFFSET)
        default_routing_statuses = _read_flags_from_word(
            int_value, DiagnosticFilterDefaultRoutingStatus,
            _DEFAULT_ROUTE_OFFSET)
        emergency_routing_statuses = _read_flags_from_word(
            int_value, DiagnosticFilterEmergencyRoutingStatus,
            _EMERGENCY_ROUTE_OFFSET)
        packet_types = _read_flags_from_word(
            int_value, DiagnosticFilterPacketType, _PACKET_TYPE_OFFSET)

        return DiagnosticFilter(
            enable_interrupt_on_counter_event,
            match_emergency_routing_status_to_incoming_packet,
            destinations, sources, payload_statuses, default_routing_statuses,
            emergency_routing_statuses, packet_types)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
from spinnman.model import CPUInfo
from spinnman.constants import CPU_INFO_BYTES
from spinnman.utilities.utility_functions import get_vcpu_address
from spinnman.messages.scp.impl import ReadMemory
from .abstract_multi_connection_process import AbstractMultiConnectionProcess


class GetCPUInfoProcess(AbstractMultiConnectionProcess):
    __slots__ = [
        "_cpu_info"]

    def __init__(self, connection_selector):
        """
        :param connection_selector:
        :type connection_selector:
            AbstractMultiConnectionProcessConnectionSelector
        """
        super().__init__(connection_selector)
        self._cpu_info = list()

    def __handle_response(self, x, y, p, response):
        self._cpu_info.append(CPUInfo(x, y, p, response.data, response.offset))

    def get_cpu_info(self, core_subsets):
        """
        :param ~spinn_machine.CoreSubsets core_subsets:
        :rtype: list(CPUInfo)
        """
        for core_subset in core_subsets:
            x = core_subset.x
            y = core_subset.y

            for p in core_subset.processor_ids:
                self._send_request(
                    ReadMemory(x, y, get_vcpu_address(p), CPU_INFO_BYTES),
                    functools.partial(self.__handle_response, x, y, p))
        self._finish()
        self.check_for_error()

        return self._cpu_info



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class SCPCommand(Enum):
    """ The SCP Commands
    """
    CMD_VER = (0, "Get SCAMP Version")
    CMD_RUN = 1
    CMD_READ = (2, "Read SDRAM")
    CMD_WRITE = (3, "Write SDRAM")
    CMD_APLX = 4
    CMD_FILL = 5
    CMD_REMAP = 16
    CMD_LINK_READ = (17, "Read neighbouring chip's memory.")
    CMD_LINK_WRITE = (18, "Write neighbouring chip's memory.")
    CMD_AR = 19
    CMD_NNP = (20, "Send a Nearest-Neighbour packet")
    CMD_P2PC = 21
    CMD_SIG = (22, "Send a Signal")
    CMD_FFD = (23, "Send Flood-Fill Data")
    CMD_AS = 24
    CMD_LED = (25, "Control the LEDs")
    CMD_IPTAG = (26, "Set an IPTAG")
    CMD_SROM = 27
    CMD_ALLOC = (28, "Allocate or Free SDRAM or Routing entries")
    CMD_RTR = (29, "Initialise the router")
    CMD_DPRI = (30, "Dropped Packet Reinjection setup")
    CMD_INFO = (31, "Get Chip Summary Information")
    CMD_BMP_INFO = (48, "Get BMP info structures")
    CMD_FLASH_COPY = 49
    CMD_FLASH_ERASE = 50
    CMD_FLASH_WRITE = 51
    CMD_RESET = 55
    CMD_BMP_POWER = (57, "Turns on or off the machine via BMP")
    CMD_TUBE = 64

    def __new__(cls, value, doc=""):
        # pylint: disable=protected-access, unused-argument
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, value, doc=""):
        self._value_ = value
        self.__doc__ = doc



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .database_confirmation import DatabaseConfirmation
from .eieio_command_header import EIEIOCommandHeader
from .eieio_command_message import EIEIOCommandMessage
from .event_stop_request import EventStopRequest
from .host_data_read import HostDataRead
from .host_data_read_ack import HostDataReadAck
from .host_send_sequenced_data import HostSendSequencedData
from .notification_protocol_pause_stop import NotificationProtocolPauseStop
from .notification_protocol_start_resume import NotificationProtocolStartResume
from .padding_request import PaddingRequest
from .spinnaker_request_buffers import SpinnakerRequestBuffers
from .spinnaker_request_read_data import SpinnakerRequestReadData
from .start_requests import StartRequests
from .stop_requests import StopRequests

__all__ = ["DatabaseConfirmation", "EIEIOCommandHeader", "EIEIOCommandMessage",
           "EventStopRequest", "HostDataRead", "HostSendSequencedData",
           "NotificationProtocolPauseStop", "NotificationProtocolStartResume",
           "PaddingRequest", "SpinnakerRequestBuffers", "HostDataReadAck",
           "SpinnakerRequestReadData", "StartRequests", "StopRequests"]



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
from spinn_utilities.overrides import overrides
from .udp_connection import UDPConnection
from .utils import update_sdp_header_for_udp_send
from spinnman.constants import SCP_SCAMP_PORT
from spinnman.messages.scp.enums import SCPResult
from spinnman.connections.abstract_classes import SCPReceiver, SCPSender

_TWO_SHORTS = struct.Struct("<2H")
_TWO_SKIP = struct.Struct("<2x")
_REPR_TEMPLATE = "BMPConnection(cabinet={}, frame={}, boards={}, " \
    "local_host={}, local_port={}, remote_host={}, remote_port={}"


class BMPConnection(UDPConnection, SCPReceiver, SCPSender):
    """ A BMP connection which supports queries to the BMP of a SpiNNaker\
        machine
    """
    __slots__ = [
        "_boards",
        "_cabinet",
        "_frame"]

    def __init__(self, connection_data):
        """
        :param BMPConnectionData connection_data:
            The description of what to connect to.
        """
        port = SCP_SCAMP_PORT if connection_data.port_num is None\
            else connection_data.port_num
        super().__init__(
            remote_host=connection_data.ip_address, remote_port=port)
        self._cabinet = connection_data.cabinet
        self._frame = connection_data.frame
        self._boards = connection_data.boards

    @property
    def cabinet(self):
        """ The cabinet ID of the BMP

        :rtype: int
        """
        return self._cabinet

    @property
    def frame(self):
        """ The frame ID of the BMP

        :rtype: int
        """
        return self._frame

    @property
    def boards(self):
        """ The set of boards supported by the BMP

        :rtype: iterable of int
        """
        return self._boards

    @property
    def chip_x(self):
        """ Defined to satisfy the SCPSender - always 0 for a BMP
        """
        return 0

    @property
    def chip_y(self):
        """ Defined to satisfy the SCPSender - always 0 for a BMP
        """
        return 0

    @overrides(SCPSender.get_scp_data)
    def get_scp_data(self, scp_request):
        update_sdp_header_for_udp_send(scp_request.sdp_header, 0, 0)
        return _TWO_SKIP.pack() + scp_request.bytestring

    @overrides(SCPReceiver.receive_scp_response)
    def receive_scp_response(self, timeout=1.0):
        data = self.receive(timeout)
        result, sequence = _TWO_SHORTS.unpack_from(data, 10)
        return SCPResult(result), sequence, data, 2

    @overrides(SCPSender.send_scp_request)
    def send_scp_request(self, scp_request):
        self.send(self.get_scp_data(scp_request))

    def __repr__(self):
        return _REPR_TEMPLATE.format(
            self._cabinet, self._frame, self._boards,
            self.local_ip_address, self.local_port, self.remote_ip_address,
            self.remote_port)



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from spinn_machine import Router
from spinnman.messages.scp.impl.fixed_route_init import FixedRouteInit
from spinnman.processes import AbstractMultiConnectionProcess


class LoadFixedRouteRoutingEntryProcess(AbstractMultiConnectionProcess):
    """ Load a fixed route routing entry onto a chip.
    """
    __slots__ = []

    def load_fixed_route(self, x, y, fixed_route, app_id=0):
        """
        :param int x: The x-coordinate of the chip, between 0 and 255;
            this is not checked due to speed restrictions.
        :param int y: The y-coordinate of the chip, between 0 and 255;
            this is not checked due to speed restrictions.
        :param ~spinn_machine.FixedRouteEntry fixed_route:
            the fixed route entry
        :param int app_id: The ID of the application with which to associate
            the routes.  If not specified, defaults to 0.
        """
        route_entry = \
            Router.convert_routing_table_entry_to_spinnaker_route(fixed_route)
        self._send_request(FixedRouteInit(x, y, route_entry, app_id))
        self._finish()
        self.check_for_error()



# Copyright (c) 2017-2019 The University of Manchester
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, BMPRequest)
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.scp import SCPRequestHeader
from .check_ok_response import CheckOKResponse

logger = FormatAdapter(logging.getLogger(__name__))


class SetPower(BMPRequest):
    """ An SCP request for the BMP to power on or power off a rack of boards
    """
    __slots__ = []

    def __init__(self, power_command, boards, delay=0.0, board_to_send_to=0):
        """
        .. note::
            There is currently a bug in the BMP that means some boards don't\
            respond to power commands not sent to BMP 0. Thus changing the\
            board_to_send_to parameter is not recommended!

        :param PowerCommand power_command: The power command being sent
        :param boards: The boards on the same backplane to power on or off
        :type boards: int or list(int)
        :param float delay:
            Number of seconds delay between power state changes of
            the different boards.
        :param int board_to_send_to:
            The optional board to send the command to if this is to be sent
            to a frame of boards.

            .. note::
                Leave this at the default because of hardware bugs.
        """

        if board_to_send_to != 0:
            logger.warning(
                "There is currently a bug in the BMP that means some boards"
                " don't respond to power commands not sent to BMP 0.  Thus "
                "changing the board_to_send_to is not recommended!")

        arg1 = (int(delay * 1000) << 16) | power_command.value
        arg2 = self.get_board_mask(boards)

        super().__init__(
            board_to_send_to,
            SCPRequestHeader(command=SCPCommand.CMD_BMP_POWER),
            argument_1=arg1, argument_2=arg2)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self):
        return CheckOKResponse("powering request", "CMD_BMP_POWER")
